from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import padding
import os, sys, platform
from binascii import a2b_base64, b2a_base64
from threading import Thread, Event
from pathlib import Path
from plyer import keystore
import secrets


class StoppableThread(Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._stop_event = Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()


def decrypt(payload, token):
     enc = a2b_base64(payload)
     IV = enc[:16]
     message = enc[16:]
     cipher = Cipher(algorithms.AES(token), modes.CBC(IV))
     decryptor = cipher.decryptor()
     data = decryptor.update(message) + decryptor.finalize()
     data = bytes((x for x in data if x >= 0x20 and x < 127))
     data = data.decode()
     return data

def encrypt(data, token):
    padder = padding.PKCS7(128).padder()
    data = padder.update(data) + padder.finalize()
    IV = os.urandom(16)
    cipher = Cipher(algorithms.AES(token), modes.CBC(IV))
    encryptor = cipher.encryptor()
    ct = encryptor.update(data) + encryptor.finalize()
    enc = IV + ct
    enc = b2a_base64(enc).decode().strip()
    return enc


def client_approval(client_id):
    True

def device_approval(device_id):
    True

if 'linux' in platform.system().lower():
    PLATFORM = 'linux'
    data_path = str(Path.home())+'/.amasdash/'
elif 'darwin' in platform.system().lower():
    PLATFORM = 'mac'
    data_path = ''
elif 'windows' in platform.system().lower():
    PLATFORM = 'windows'
    data_path = str(Path.home())+'/AppData/Local/AMASDash/'
else:
    PLATFORM = 'other'
    data_path = ''

if 'aarch' in platform.machine().lower() and (PLATFORM == 'linux' or PLATFORM == 'windows') and sys.maxsize > 2**32:
    ARCH = 'aarch64'
elif ('amd' in platform.machine().lower() or 'x86' in platform.machine().lower()) and (PLATFORM == 'linux' or PLATFORM == 'windows') and sys.maxsize > 2**32:
    ARCH = 'x86_64'
elif PLATFORM == 'other' or PLATFORM == 'mac':
    ARCH = ''
else:
    sys.exit()

base_dir = str(os.getcwd())
if 'assets' in base_dir:
    base_dir = base_dir[:-6]
else:
    base_dir = base_dir + '/'

interpreter_path = str(sys.executable)
if '.exe' in interpreter_path:
    interpreter_path = interpreter_path[:-4]
if 'AMASDash' in interpreter_path:
    base_dir = interpreter_path[:-8]
    print('Running with pre-compiled Python')
else:
    data_path = ''
    print('Running in developer mode')

if PLATFORM == 'windows':
    if os.path.isfile(base_dir+'portable'):
        MODE = 'portable'
    else:
        MODE = 'installer'
elif PLATFORM == 'linux':
    if os.path.isfile(base_dir+'portable'):
        MODE = 'portable'
    elif os.path.isfile(base_dir+'appimage'):
        MODE = 'appimage'
    elif os.path.isfile(base_dir+'flatpak'):
        MODE = 'flatpak'
    elif os.path.isfile(base_dir+'rpm'):
        MODE = 'rpm'
    elif os.path.isfile(base_dir+'deb'):
        MODE = 'deb'
    else:
        MODE = 'installer'
elif PLATFORM == 'other' or PLATFORM == 'mac':
    MODE = 'installer'

if data_path:
    print('Using '+MODE+' version')

try:
    base64_key = keystore.get_key(servicename='AMASDash', key='base64key')
    if not base64_key:
        raise Exception
    key = a2b_base64(base64_key)
    print('Got existing encryption key!')
    keystore.set_key(servicename='AMASDash', key='base64key', value=base64_key)
except:
    try:
        os.remove(data_path+'locals.json')
    except: pass
    try:
        os.remove(data_path+'shadows.json')
    except: pass
    try:
        os.remove(data_path+'user.json')
    except:pass
    try:
        with open(data_path+'cookies.json') as cookie_file:
            if not 'local_only' in cookie_file.read():
                tmp = True
            else:
                tmp = False
        if tmp:
            os.remove(data_path+'cookies.json')
    except:
        pass
    key = secrets.token_bytes(16)
    base64_key = b2a_base64(key)
    keystore.set_key(servicename='AMASDash', key='base64key', value=base64_key.strip().decode())
    print('Created new encryption key!')