 # coding: utf-8

from kivy.storage.jsonstore import JsonStore
from kivymd.uix.screen import MDScreen
from kivymd.uix.boxlayout import MDBoxLayout
from kivymd.uix.button import MDButton, MDIconButton, MDButtonText, MDButtonIcon
from kivymd.uix.card import MDCard
from kivymd.uix.label import MDLabel
from kivymd.uix.floatlayout import MDFloatLayout
from kivymd.uix.scrollview import MDScrollView
from kivymd.uix.selectioncontrol import MDCheckbox
from kivymd.uix.menu import MDDropdownMenu
from kivymd.uix.progressindicator import MDCircularProgressIndicator
from kivymd.uix.dialog import (
    MDDialog,
    MDDialogIcon,
    MDDialogHeadlineText,
    MDDialogSupportingText,
    MDDialogButtonContainer,
    MDDialogContentContainer,
)
from kivymd.uix.divider import MDDivider
from kivymd.uix.textfield import (
    MDTextField,
    MDTextFieldLeadingIcon,
    MDTextFieldHintText,
    MDTextFieldHelperText,
)
from kivymd.uix.segmentedbutton import (
    MDSegmentedButton,
    MDSegmentedButtonItem,
    MDSegmentButtonLabel,
)
from kivymd.uix.pickers import MDTimePickerInput
from kivy.uix.modalview import ModalView
from kivy.uix.image import Image
from kivy.uix.widget import Widget
from kivy.core.window import Window
from kivy.clock import Clock, mainthread
from threading import Thread, current_thread
from datetime import datetime, UTC
from time import sleep
import os, shutil
from json import dumps, loads
import requests
import asyncio
import aiohttp
from paho.mqtt.packettypes import PacketTypes
import ssl, certifi
import socket
from paho.mqtt.properties import Properties
import paho.mqtt.client as mqtt
from urllib.parse import quote
import webbrowser
import time


from assets.env import *


import warnings
warnings.showwarning = print

wait_for_resize_refresh = True
show_advanced = False
    

class DashboardScreen(MDScreen):
    def __init__(self, **kwargs):
        super(DashboardScreen, self).__init__(**kwargs)

        Window.bind(on_resize=self.on_window_resize)

        self.loading_dialog = ModalView(
            auto_dismiss=False,
            size_hint=(1,1),
            background_color=(0,0,0,0)
        )
        self.loading_dialog.add_widget(
            MDFloatLayout(
                MDCircularProgressIndicator(
                    size_hint=(None,None),
                    size=('24dp','24dp'),
                    pos_hint={'center_x': 0.5, 'center_y': 0.5},
                )
            ),
        )

        self.account_menu_items = [
            {
                "text": 'Login/Signup',
                "on_release": lambda *args: self.login(),
            },
            {
                "text": 'Logout (keep local devices)',
                "on_release": lambda *args: self.load_logout_only(),
            },
            {
                "text": 'Logout (delete local devices)',
                "on_release": lambda *args: self.load_logout_del_device(),
            },
        ]

        self.hamburger_menu_items = [
            {
                "text": 'Dashboard',
                "on_release": lambda *args: self.redirect_screen('dashboard'),
            },
            {
                "text": 'Setup a new device',
                "on_release": lambda *args: self.redirect_screen('setup'),
            },
            {
                "text": 'Add existing local device',
                "on_release": lambda *args: self.redirect_screen('local_add'),
            },
            {
                "text": 'Manual Update',
                "on_release": lambda *args: self.redirect_screen('update'),
            },
            {
                "text": 'Switch Dark/Light Mode',
                "on_release": lambda *args: self.change_display_mode(),
            },
            {
                "text": 'About',
                "on_release": lambda *args: self.redirect_screen('about'),
            },
        ]

        self.settings_store = JsonStore(data_path+'settings.json')
        self.local_store = JsonStore(data_path+'locals.json')
        self.shadow_store = JsonStore(data_path+'shadows.json')
        self.cookie_jar = JsonStore(data_path+'cookies.json')
        self.user_store = JsonStore(data_path+'user.json')

        if 'refresh_token' in self.cookie_jar.keys():
            del self.account_menu_items[0]
        elif 'local_only' in self.cookie_jar.keys():
            del self.account_menu_items[-2]

        self.account_menu = MDDropdownMenu(
            caller=self.ids.account_button,
            items=self.account_menu_items,
        )

        self.hamburger_menu = MDDropdownMenu(
            caller=self.ids.hamburger_button,
            items=self.hamburger_menu_items,
        )

        self.local_ips = {}
        self.api_keys = {}
        self.local_only = []
        self.locally_available = []
        self.locally_available_errored = []
        self.local_only_unavailable = []
        self.remotely_available = []
        self.remote_unavail = []
        self.user_interactions = []
        self.toggle_interactions = {}
        self.local_ws_connections = {}
        self.mqtt_sub = None

        self._coordinator_thread = None
        self._local_data_thread = None
        self._local_data_loop = None
        self._remote_data_thread = None
        self._new_device_discovery_thread = None
        self._update_check_thread = None
        self._internet = Event()
        self._internet_thread = StoppableThread(target=self.check_connection, args=([self._internet]), daemon=True, name='internet-check')
        self._internet_thread.start()

        Thread(target=self.first_run_tasks, daemon=True, name='first-run').start()


    @mainthread
    def show_loading(self):
        try:
            self.loading_dialog.open()
        except Exception as e:
            print(e)


    @mainthread
    def dismiss_loading(self):
        try:
            self.loading_dialog.dismiss()
        except Exception as e:
            print(e)


    @mainthread
    def refresh_menu_items(self):
        self.account_menu_items = [
            {
                "text": 'Login/Signup',
                "on_release": lambda *args: self.login(),
            },
            {
                "text": 'Logout (keep local devices)',
                "on_release": lambda *args: self.load_logout_only(),
            },
            {
                "text": 'Logout (delete local devices)',
                "on_release": lambda *args: self.load_logout_del_device(),
            },
        ]
        if 'refresh_token' in self.cookie_jar.keys():
            del self.account_menu_items[0]
        elif 'local_only' in self.cookie_jar.keys():
            del self.account_menu_items[-2]
        self.account_menu = MDDropdownMenu(
            caller=self.ids.account_button,
            items=self.account_menu_items,
        )

    
    def on_window_resize(self, window, width, height):
        global wait_for_resize_refresh

        print('Width: '+str(width)+', Height: '+str(height))
        if width < 600:
            self.settings_store.put('modal_width_hint', value=0.9)
        elif width < 800:
            self.settings_store.put('modal_width_hint', value=0.8)
        else:
            self.settings_store.put('modal_width_hint', value=0.7)

        if not wait_for_resize_refresh: self.refresh_entire_window()

    
    def check_connection(self, internet_on: Event, interval: float = 60.0) -> None:
        # print("starting function intended to be in a new thread")
        internet_on.clear()
        # one could also import Queue from queue and return specific error
        # messages by putting them in the queue in this thread and checking
        # for them in the main thread.
        while True:
            try:
                _host = socket.gethostbyname('one.one.one.one')
                s = socket.create_connection((_host, 80), 2)
                print('Checking internet connection!')
            except Exception as e:
                print('Internet is NOT connected!')
                print(e)
                internet_on.clear()
            else:
                print('Internet is connected!')
                internet_on.set()
                s.close()
            sleep(interval)


    def on_touch_down(self, touch):
        # get touch position in coordinates of the main_box (parent of inner_scroll)
        x,y = self.ids.parent_box.to_widget(*self.to_window(*touch.pos))

        # if converted position is within the inner_scroll, send touch to the inner_scroll
        if self.ids.dash_inner_scroll.collide_point(x,y):
            touch.pos = (x,y)   # change touch position to coordinates in main_box
            return self.ids.dash_inner_scroll.on_touch_down(touch)   # call on_touch of inner_scroll
        else:
            return super(DashboardScreen, self).on_touch_down(touch)


    @mainthread
    def redirect_screen(self, screen):
        self.hamburger_menu.dismiss()
        self.manager.current = screen


    def load_login(self):
        self.show_loading()
        Thread(target=self.login, daemon=True, name='login').start()
    

    def login(self):
        self.cookie_jar = JsonStore(data_path+'cookies.json')
        if 'local_only' in self.cookie_jar.keys():
            self.cookie_jar.delete('local_only')
        settings_store = JsonStore(data_path+'settings.json')
        flask_host = settings_store.get('flask')['host']
        flask_port = settings_store.get('flask')['port']
        webbrowser.open('http://'+flask_host+':'+str(flask_port))
        self.cookie_jar = JsonStore(data_path+'cookies.json')
        self.cookie_jar.put('wait_for_validation', value=True)
        while True:
            self.cookie_jar = JsonStore(data_path+'cookies.json')
            if self.cookie_jar.count() and 'wait_for_validation' not in self.cookie_jar.keys():
                break
            else:
                print('Waiting for cookies to be baked!')
                sleep(2)
        user_t = {}
        if self.cookie_jar.count():
            for i in self.cookie_jar.keys():
                if i not in ['wait_for_validation', 'local_only']:
                    user_t[i] = decrypt(self.cookie_jar.get(i)['value'], key)
        self.initialize({'recheck': True}, user_t, True, True)
        self.account_menu.dismiss()
        self.refresh_menu_items()
        self.dismiss_loading()
        self.get_availability_lists()
        self.refresh_entire_window()


    def change_display_mode(self):
        self.theme_cls.theme_style = (
            "Dark" if self.theme_cls.theme_style == "Light" else "Light"
        )
        self.settings_store.put('theme', value=self.theme_cls.theme_style)
        self.refresh_entire_window()


    def load_diff(self, old, new):
        for i in new:
            if type(new[i]) is dict:
                for j in new[i]:
                    if type(new[i][j]) is dict:
                        for k in new[i][j]:
                            if new[i][j][k] != old[i][j][k]:
                                old[i][j][k]= new[i][j][k]
                    else:
                        if new[i][j] != old[i][j]:
                            old[i][j]= new[i][j]
            else:
                if new[i] != old[i]:
                    old[i]= new[i]


    @mainthread
    def control_modal(self, device_id):
        global show_advanced

        if device_id in self.locally_available:
            try:
                state = loads(decrypt(self.local_store.get(device_id)['value'], key))
            except:
                state = loads(decrypt(self.shadow_store.get(device_id)['value'], key))
        else:
            state = loads(decrypt(self.shadow_store.get(device_id)['value'], key))

        desired_state = {}

        
        def utc_to_local(value):
            return datetime.strptime(datetime.today().astimezone(UTC).strftime('%Y-%m-%d') + ' ' + value, '%Y-%m-%d %H%M').replace(tzinfo=UTC).astimezone().strftime('%H:%M')

        
        def local_to_utc(value):
            return datetime.strptime(datetime.today().strftime('%Y-%m-%d') + ' '+value, '%Y-%m-%d %H:%M').astimezone(UTC).strftime('%H%M')

        
        def on_time_ok(time_picker_vertical: MDTimePickerInput, item: str, view):
            nonlocal desired_state, state

            desired_time = str(time_picker_vertical.time.hour) + ':' + str(time_picker_vertical.time.minute)
            if 'light' in desired_state:
                desired_state['light'][item] = local_to_utc(desired_time)
            else:
                desired_state['light'] = {}
                desired_state['light'][item] = local_to_utc(desired_time)
            state['state']['reported']['light'][item] = local_to_utc(desired_time)
            view.dismiss()
            time_picker_vertical.dismiss()
            update_modal(state)

            
        def on_time_cancel(time_picker: MDTimePickerInput):
            time_picker.dismiss()

        
        def state_to_card_time(item):
            nonlocal state

            print('Converting state items to card items')

            if utc_to_local(state['state']['reported']['light'][item]).split(':')[0] == '00':
                card_time = ['12', utc_to_local(state['state']['reported']['light'][item]).split(':')[1], 'AM']
            elif int(utc_to_local(state['state']['reported']['light'][item]).split(':')[0]) <= 11:
                card_time = [utc_to_local(state['state']['reported']['light'][item]).split(':')[0], utc_to_local(state['state']['reported']['light'][item]).split(':')[1], 'AM']
            elif utc_to_local(state['state']['reported']['light'][item]).split(':')[0] == '12':
                card_time = ['12', utc_to_local(state['state']['reported']['light'][item]).split(':')[1], 'PM']
            else:
                card_time = [str(int(utc_to_local(state['state']['reported']['light'][item]).split(':')[0])-12) if len(str(int(utc_to_local(state['state']['reported']['light'][item]).split(':')[0])-12)) == 2 else '0'+str(int(utc_to_local(state['state']['reported']['light'][item]).split(':')[0])-12), utc_to_local(state['state']['reported']['light'][item]).split(':')[1], 'PM']
            card_time_str = card_time[0]+':'+card_time[1]+' '+card_time[2]
            return card_time, card_time_str
        
        
        def light_schedule_picker(item, view):

            card_time = state_to_card_time(item)[0]
            time_picker = MDTimePickerInput(hour=card_time[0], minute=card_time[1], am_pm=card_time[2].lower())
            time_picker.bind(on_ok=lambda *args: on_time_ok(time_picker, item, view))
            time_picker.bind(on_cancel=on_time_cancel)
            time_picker.open()
        

        def on_dialog_ok(dialog_modal: MDDialog):
            dialog_modal.dismiss()
        

        def load_on_dialog_regenerate(info: str, state: dict, dialog_modal: MDDialog, control_modal: ModalView):
            self.show_loading()
            Thread(target=lambda *args: on_dialog_regenerate(info, state, dialog_modal, control_modal), daemon=True, name='regenerate-'+info).start()


        def on_dialog_regenerate(info: str, state: dict, dialog_modal: MDDialog, control_modal: ModalView):
            nonlocal self

            if info == 'api_key':
                try:
                    resp = requests.post('https://'+state['state']['reported']['internal_ip']+'/api_key', headers={'x-api-key': state['state']['reported']['api_key'], 'Content-Type': 'application/json', 'Accept': 'application/json'}, json={"generate": True}, verify=base_dir+'assets/static/AMAS_CA.pem')
                    if resp:
                        new_api_key = resp.json()
                        new_api_key = new_api_key['state']['new']['api_key']
                        local_store = JsonStore(data_path+'locals.json')
                        existing_data = loads(decrypt(local_store.get(state['state']['reported']['device_id'])['value'], key))
                        existing_data['state']['reported']['api_key'] = new_api_key
                        local_store.put(state['state']['reported']['device_id'], value=encrypt(dumps(existing_data).encode(), key))
                        self.api_keys[state['state']['reported']['device_id']] = new_api_key
                        shadow_store = JsonStore(data_path+'shadows.json')
                        if state['state']['reported']['device_id'] in shadow_store.keys():
                            existing_shadow = loads(decrypt(shadow_store.get(state['state']['reported']['device_id'])['value'], key))
                            existing_shadow['state']['reported']['api_key'] = new_api_key
                            shadow_store.put(state['state']['reported']['device_id'], value=encrypt(dumps(existing_shadow).encode(), key))
                        dialog_modal.dismiss()
                        control_modal.dismiss()
                except:
                    pass
            elif info == 'salt':
                try:
                    resp = requests.post('https://'+state['state']['reported']['internal_ip']+'/salt', headers={'x-api-key': state['state']['reported']['api_key'], 'Content-Type': 'application/json', 'Accept': 'application/json'}, json={"generate": True}, verify=base_dir+'assets/static/AMAS_CA.pem')
                    if resp:
                        dialog_modal.dismiss()
                        control_modal.dismiss()
                except:
                    pass

            self.dismiss_loading()

        
        def dialog_info(info: str, state: dict, view: ModalView):
            good_request = False
            if info == 'device_id':
                headline = 'Device ID'
                text_helper = 'This identifier uniquely identifies this specific device in AMAS backend services.'
                good_request = True
            elif info == 'api_key':
                headline = 'API Key'
                text_helper = 'This key authenticates any requests to your device. Without this key, you will not be able to get data from, or send instructions to your device when connected locally. This can be regenerated, but regenerating your API key breaks all existing local connections other than this app.'
                good_request = True
            elif info == 'salt':
                headline = 'Salt'
                text_helper = 'This random string of characters anonymizes your information on AMAS data collection service (if enabled). This information does not leave your local network, and can be regenerated at any time.'
                good_request = True

            if good_request:
                dialog_modal = MDDialog(
                    MDDialogIcon(
                        icon='hammer-screwdriver',
                    ),
                    MDDialogHeadlineText(
                        text=headline,
                    ),
                    MDDialogSupportingText(
                        text=text_helper,
                    ),
                    MDDialogContentContainer(MDBoxLayout(
                        MDDivider(),
                        MDBoxLayout(
                            size_hint_y=None,
                            height='15dp',
                        ),
                        MDLabel(
                            text=state['state']['reported'][info] if info != 'salt' else state['state']['admin'][info],
                            adaptive_size=True,
                            allow_selection=True,
                            allow_copy=True,
                            pos_hint={'center_x': 0.5},
                        ),
                        MDBoxLayout(
                            size_hint_y=None,
                            height='15dp',
                        ),
                        MDDivider(),
                        size_hint_y=None,
                        height='55dp',
                        orientation='vertical',
                    )),
                )
                if info not in ['api_key', 'salt']:
                    dialog_modal.add_widget(
                        MDDialogButtonContainer(
                            Widget(),
                            MDButton(
                                MDButtonText(text="OK"),
                                style="text",
                                on_release=lambda *args: on_dialog_ok(dialog_modal),
                            )
                        ),
                    )
                else:
                    dialog_modal.add_widget(
                        MDDialogButtonContainer(
                            Widget(),
                            MDButton(
                                MDButtonText(text="Regenerate"),
                                style="text",
                                on_release=lambda *args: load_on_dialog_regenerate(info, state, dialog_modal, view),
                            ),
                            MDButton(
                                MDButtonText(text="OK"),
                                style="text",
                                on_release=lambda *args: on_dialog_ok(dialog_modal),
                            ),
                            spacing="8dp",
                        ),
                    )
                dialog_modal.open()


        def dev_cert_state_to_card(shadow):
            if 'device_certificate' in shadow['state']['reported']:
                if shadow['state']['reported']['device_certificate']:
                    if shadow['state']['reported']['device_certificate']['renewal_period_active']:
                        dev_cert = 'Renewal Needed'
                    else:
                        dev_cert = 'Installed'
                else:
                    dev_cert = 'Not Installed'
            else:
                dev_cert = 'Not Registered'
            return dev_cert


        def control_device(device_id: str, raw_desired_state: dict, modal_box: MDBoxLayout, view: ModalView, screen=self):
            print('Got device ID: '+device_id)
            print('Got existing desired state: ')
            desired_state = {'state': {}}
            desired_state['state']['desired'] = raw_desired_state
            print(desired_state)
            local_store = JsonStore(data_path+'locals.json')
            shadow_store = JsonStore(data_path+'shadows.json')
            if device_id in screen.locally_available:
                try:
                    existing_state = loads(decrypt(local_store.get(device_id)['value'], key))
                except:
                    existing_state = loads(decrypt(shadow_store.get(device_id)['value'], key))
            else:
                existing_state = loads(decrypt(shadow_store.get(device_id)['value'], key))
            print('Loaded existing reported state')
            
            controllable_floats = ['factory_reset_float', 'network_reset_float', 'del_device_float', 'debug_float', 'anon_data_float', 'auto_update_float', 'pump_drain_float', 'pump_interval_float', 'pump_run_float', 'pump_power_float', 'light_over_float']
            for i in modal_box.children:
                if i.id in controllable_floats:
                    print(i.id)
                    for j in i.children:
                        if j.id:
                            print(j.id)
                            if 'factory_reset_' in j.id:
                                if j.state == 'down':
                                    print('Sending factory reset instructions')
                                    if 'admin' in desired_state['state']:
                                        desired_state['state']['admin']['factory_reset'] = True
                                    else:
                                        desired_state['state']['admin'] = {}
                                        desired_state['state']['admin']['factory_reset'] = True
                            if 'network_reset_' in j.id:
                                if j.state == 'down':
                                    print('Sending network reset instructions!')
                                    if 'admin' in desired_state['state']:
                                        desired_state['state']['admin']['network_reset'] = True
                                    else:
                                        desired_state['state']['admin'] = {}
                                        desired_state['state']['admin']['network_reset'] = True
                            if 'del_device_' in j.id:
                                if j.state == 'down':
                                    print('Not implemented yet!')
                            if 'pump_run_' in j.id:
                                if j.text:
                                    if 'pump' in desired_state['state']['desired']:
                                        desired_state['state']['desired']['pump']['runtime'] = int(j.text)
                                    else:
                                        desired_state['state']['desired']['pump'] = {}
                                        desired_state['state']['desired']['pump']['runtime'] = int(j.text)
                                else:
                                    print('Pump runtime not changed!')
                            elif 'pump_interval_' in j.id:
                                if j.text:
                                    if 'pump' in desired_state['state']['desired']:
                                        desired_state['state']['desired']['pump']['interval'] = int(j.text)
                                    else:
                                        desired_state['state']['desired']['pump'] = {}
                                        desired_state['state']['desired']['pump']['interval'] = int(j.text)
                                else:
                                    print('Pump interval not changed!')
                            elif 'debug_' in j.id:
                                print('Existing debug: ')
                                print(existing_state['state']['admin']['debug'])
                                try:
                                    if existing_state['state']['admin']['debug'] != j.state == 'down':
                                        if 'admin' in desired_state['state']:
                                            desired_state['state']['admin']['debug'] = j.state == 'down'
                                        else:
                                            desired_state['state']['admin'] = {}
                                            desired_state['state']['admin']['debug'] = j.state == 'down'
                                    else:
                                        raise Exception
                                except:
                                    print('Debug state not changed!')
                            elif 'auto_update_' in j.id:
                                print('Existing auto update: ')
                                print(existing_state['state']['reported']['auto_update'])
                                if existing_state['state']['reported']['auto_update'] != j.state == 'down':
                                    desired_state['state']['desired']['auto_update'] = j.state == 'down'
                                else:
                                    print('Auto update state not changed!')
                            elif 'anon_data_' in j.id:
                                print('Existing anonymous data collection: ')
                                try:
                                    print(existing_state['state']['reported']['anonymous_data_collection'])
                                except:
                                    print(True)
                                try:
                                    if existing_state['state']['reported']['anonymous_data_collection'] != j.state == 'down':
                                        desired_state['state']['desired']['anonymous_data_collection'] = j.state == 'down'
                                    else:
                                        print('Anonymous data collection not changed!')
                                except:
                                    print('Anonymous data collection not changed!')
                            elif 'pump_drain_' in j.id:
                                print('Existing pump drain: ')
                                print(existing_state['state']['reported']['pump']['drain'])
                                if existing_state['state']['reported']['pump']['drain'] != j.state == 'down':
                                    if 'pump' in desired_state['state']['desired']:
                                        desired_state['state']['desired']['pump']['drain'] = j.state == 'down'
                                    else:
                                        desired_state['state']['desired']['pump'] = {}
                                        desired_state['state']['desired']['pump']['drain'] = j.state == 'down'
                                else:
                                    print('Drain status not changed!')
                            elif 'pump_power_' in j.id:
                                print('Existing pump power: ')
                                print(existing_state['state']['reported']['pump']['powered'])
                                if existing_state['state']['reported']['pump']['powered'] != j.state == 'down':
                                    if 'pump' in desired_state['state']['desired']:
                                        desired_state['state']['desired']['pump']['powered'] = j.state == 'down'
                                    else:
                                        desired_state['state']['desired']['pump'] = {}
                                        desired_state['state']['desired']['pump']['powered'] = j.state == 'down'
                                else:
                                    print('Pump powered status not changed!')
                            elif 'light_over_' in j.id:
                                k = j.get_marked_items()[0]
                                print(k.id)
                                ins = k.id.split('_')[2]
                                
                                if ins == 'on':
                                    ins = '1'
                                elif ins == 'off':
                                    ins = '0'
                                else:
                                    ins = False

                                if existing_state['state']['reported']['light']['override'] != ins:
                                    if 'light' in desired_state['state']['desired']:
                                        desired_state['state']['desired']['light']['override'] = ins
                                    else:
                                        desired_state['state']['desired']['light'] = {}
                                        desired_state['state']['desired']['light']['override'] = ins
                                else:
                                    print('Light override status not changed!')
                else:
                    if i.id:
                        print(i.id)
            print('Built final desired json: ')
            print(desired_state)
            print('Sending instructions to: '+device_id)
            if device_id in screen.locally_available:
                try:
                    resp = requests.post('https://'+screen.local_ips[device_id]['ip']+'/control', headers={'x-api-key': screen.api_keys[device_id], 'Content-Type': 'application/json', 'Accept': 'application/json'}, json=desired_state, verify=base_dir+'assets/static/AMAS_CA.pem')
                    print('Got response')
                    if resp:
                        local_store.put(device_id, value=encrypt(dumps(resp.json()).encode(), key))
                        print('Saved new data')
                        view.dismiss()
                        screen.refresh_single_card(device_id)
                    else:
                        raise Exception
                except Exception as e:
                    print(e)
            elif device_id in screen.remotely_available:
                user_store = JsonStore(data_path+'user.json')
                user_c = {}
                for i in user_store.keys():
                    user_c[i] = screen.format_certs(decrypt(user_store.get(i)['value'], key))
                topic = '$aws/things/' + user_c['thingName'] + '/shadow/name/' + device_id + '/update'
                topic = quote(topic)
                publish_url = 'https://iotcore.amastech.cloud:8443/topics/' + topic + '?qos=1'
                if not os.path.exists(data_path+'.tmp'):
                    os.makedirs(data_path+'.tmp')
                if not os.path.exists(data_path+'.tmp/'+'mqtt_client_certs'):
                    os.makedirs(data_path+'.tmp/'+'mqtt_client_certs')
                with open(data_path+'.tmp/'+'mqtt_client_certs'+'/cert.pem', 'w') as cert_file:
                    cert_file.write(user_c['cert'])
                    certfile = data_path+'.tmp/'+'mqtt_client_certs'+'/cert.pem'
                with open(data_path+'.tmp/'+'mqtt_client_certs'+'/key.pem', 'w') as key_file:
                    key_file.write(user_c['key'])
                    keyfile = data_path+'.tmp/'+'mqtt_client_certs'+'/key.pem'
                print('Sending instructions to IoT broker')
                try:
                    resp = requests.post(publish_url, json=desired_state, cert=(certfile, keyfile), verify=certifi.where())
                    if resp.status_code == 200:
                        print('Request successful!')
                        Clock.schedule_once(screen.refresh_entire_window)
                        view.dismiss()
                    else:
                        print('Request failed!')
                        print(resp.content)
                except:
                    print('Request failed!')
                try:
                    shutil.rmtree(data_path+'.tmp/'+'mqtt_client_certs', ignore_errors=True)
                except:
                    pass
            screen.dismiss_loading()

        
        def load_control_device(device_id: str, raw_desired_state: dict, modal_box: MDBoxLayout, view: ModalView, screen=self):
            screen.show_loading()
            Thread(target=lambda *args: control_device(device_id, raw_desired_state, modal_box, view), daemon=True, name='control-'+device_id).start()


        def show_advanced_toggle(view: ModalView):
            global show_advanced
            nonlocal state

            show_advanced = not show_advanced
            
            view.dismiss()
            update_modal(state)

        def update_modal(state):
            global show_advanced

            try:
                width_hint = self.settings_store.get('modal_width_hint')['value']
            except:
                try:
                    width = Window.size[0]
                    if width < 600:
                        self.settings_store.put('modal_width_hint', value=0.9)
                        width_hint = 0.9
                    elif width < 800:
                        self.settings_store.put('modal_width_hint', value=0.8)
                        width_hint = 0.8
                    else:
                        self.settings_store.put('modal_width_hint', value=0.7)
                        width_hint = 0.7
                except:
                    width_hint = 0.8
            device_id = state['state']['reported']['device_id']
            if width_hint:
                view = ModalView(size_hint=(width_hint, 0.7), border=(30, 30, 30, 30),)
            else:
                view = ModalView(size_hint=(0.8, 0.7), border=(30, 30, 30, 30),)
            try:
                anon_data_coll = state['state']['reported']['anonymous_data_collection']
            except:
                anon_data_coll = True
            modal_scroll = MDScrollView()
            modal_box_height = 1100
            modal_box = MDBoxLayout(
                    MDFloatLayout(Image(
                        source=base_dir+'assets/icons/AMAS.png',
                        height="60dp",
                        size_hint_y=None,
                        pos_hint={"center_x": 0.5, "center_y": 0.5}
                    ),
                    MDIconButton(
                        icon='cog-off' if show_advanced else 'cog',
                        pos_hint={"right": 1, "center_y": 0.5},
                        on_release=lambda *args: show_advanced_toggle(view),
                    )),
                    MDBoxLayout(
                        height="15dp",
                        size_hint_y=None,
                    ),
                    MDFloatLayout(MDTextField(
                        MDTextFieldHintText(text=state['state']['reported']['friendly_name'] if 'friendly_name' in state['state']['reported'] else state['state']['reported']['serial']),
                        MDTextFieldHelperText(text='Enter friendly name (optional)'),
                        pos_hint={'center_x': 0.5, 'top': 1},
                        id='friendly_name_'+device_id,
                        mode="outlined",
                    )),
                    MDFloatLayout(MDLabel(
                        text='Light',
                        adaptive_size=True,
                        font_style='Title',
                        role='large',
                        padding='5dp',
                        pos_hint={'center_x': 0.5, 'top': 1},
                    )),
                    MDFloatLayout(
                        MDLabel(
                            text='On: ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                        MDButton(
                            MDButtonIcon(icon='clock-edit'),
                            MDButtonText(text=state_to_card_time('on')[1]),
                            style='text',
                            pos_hint={'right': 1, 'center_y': .5},
                            on_release=lambda *args: light_schedule_picker('on', view)
                        ),
                    ),
                    MDFloatLayout(
                        MDLabel(
                            text='Off: ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                        MDButton(
                            MDButtonIcon(icon='clock-edit'),
                            MDButtonText(text=state_to_card_time('off')[1]),
                            style='text',
                            pos_hint={'right': 1, 'center_y': .5},
                            on_release=lambda *args: light_schedule_picker('off', view)
                        ),
                    ),
                    MDFloatLayout(
                        MDLabel(
                            text='Schedule Override: ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5 if Window.size[0]>555 else 0.8},
                            padding='30dp',
                        ),
                        MDSegmentedButton(
                            MDSegmentedButtonItem(
                                MDSegmentButtonLabel(text='False'),
                                id='light_over_false_'+device_id,
                                active=True if not state['state']['reported']['light']['override'] else False,
                                ),
                            MDSegmentedButtonItem(
                                MDSegmentButtonLabel(text='Force On'),
                                id='light_over_on_'+device_id,
                                active=True if state['state']['reported']['light']['override'] and state['state']['reported']['light']['override'] == '1' else False,
                                ),
                            MDSegmentedButtonItem(
                                MDSegmentButtonLabel(text='Force Off'),
                                id='light_over_off_'+device_id,
                                active=True if state['state']['reported']['light']['override'] and state['state']['reported']['light']['override'] == '0' else False,
                                ),
                            pos_hint={'right': 1, 'center_y': 0.5 if Window.size[0]>555 else 0.3},
                            size_hint_x=0.3,
                            spacing='5dp',
                            adaptive_width=True,
                            type='small',
                            id='light_over_button_'+device_id,
                        ),
                        id='light_over_float'
                    ),
                    MDFloatLayout(MDLabel(
                        text='Pump',
                        adaptive_size=True,
                        font_style='Title',
                        role='large',
                        padding='5dp',
                        pos_hint={'center_x': 0.5, 'top': 1},
                    )),
                    MDFloatLayout(
                        MDLabel(
                            text='Scheduled Run: ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                        MDCheckbox(
                            pos_hint={'right': 1, 'center_y': .5},
                            padding='40dp',
                            id='pump_power_'+device_id,
                            active=state['state']['reported']['pump']['powered'],
                        ),
                        id='pump_power_float',
                    ),
                    MDFloatLayout(
                        MDLabel(
                            text='Runtime (seconds): ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                        MDTextField(
                            MDTextFieldLeadingIcon(icon='timer'),
                            MDTextFieldHintText(text=str(state['state']['reported']['pump']['runtime'])),
                            MDTextFieldHelperText(text='60 - 7200'),
                            pos_hint={'right': 1, 'center_y': .5},
                            id='pump_run_'+device_id,
                            input_filter=int,
                            size_hint_x=None,
                            width="100dp",
                            mode="outlined",
                        ),
                        id='pump_run_float',
                    ),
                    MDFloatLayout(
                        MDLabel(
                            text='Interval (seconds): ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                        MDTextField(
                            MDTextFieldLeadingIcon(icon='timer'),
                            MDTextFieldHintText(text=str(state['state']['reported']['pump']['interval'])),
                            MDTextFieldHelperText(text='600 - 86400'),
                            pos_hint={'right': 1, 'center_y': .5},
                            id='pump_interval_'+device_id,
                            input_filter=int,
                            size_hint_x=None,
                            width="100dp",
                            mode="outlined",
                        ),
                        id='pump_interval_float',
                    ),
                    MDFloatLayout(
                        MDLabel(
                            text='Drain: ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                        MDCheckbox(
                            pos_hint={'right': 1, 'center_y': .5},
                            padding='40dp',
                            id='pump_drain_'+device_id,
                            active=state['state']['reported']['pump']['drain'],
                        ),
                        id='pump_drain_float',
                    ),
                    orientation="vertical",
                    pos_hint={'center_x': 0.5, 'center_y': .5},
                    size_hint_y=None,
                    height=str(modal_box_height)+'dp',
                    padding='25dp',
                    md_bg_color=self.theme_cls.backgroundColor,
            )
            modal_box.add_widget(MDFloatLayout(MDLabel(
                        text='Device Properties',
                        adaptive_size=True,
                        font_style='Title',
                        role='large',
                        padding='5dp',
                        pos_hint={'center_x': 0.5, 'top': 1},
                    )),
            )
            if show_advanced:
                modal_box.add_widget(MDFloatLayout(
                        MDLabel(
                            text='Auto Update: ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                        MDCheckbox(
                            pos_hint={'right': 1, 'center_y': .5},
                            padding='40dp',
                            id='auto_update_'+device_id,
                            active=state['state']['reported']['auto_update'],
                        ),
                        id='auto_update_float',
                    ),
                )
            modal_box.add_widget(MDFloatLayout(
                        MDLabel(
                            text='Internal IP: ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                        MDLabel(
                            text=state['state']['reported']['internal_ip'],
                            adaptive_size=True,
                            pos_hint={'right': 1, 'center_y': 0.5},
                            padding='30dp',
                            allow_selection=True,
                            allow_copy=True,
                        ),
                    ),
            )
            modal_box.add_widget(MDFloatLayout(
                        MDLabel(
                            text='Registered: ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                        MDLabel(
                            text='Yes' if 'customer_id' in state['state']['reported'] else 'No',
                            adaptive_size=True,
                            pos_hint={'right': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                    ),
            )
            modal_box.add_widget(MDFloatLayout(
                        MDLabel(
                            text='Device Certificate: ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                        MDLabel(
                            text=dev_cert_state_to_card(state),
                            adaptive_size=True,
                            pos_hint={'right': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                    ),
            )
            if show_advanced:
                modal_box.add_widget(MDFloatLayout(
                        MDLabel(
                            text='Serial: ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                        MDLabel(
                            text=state['state']['reported']['serial'],
                            adaptive_size=True,
                            pos_hint={'right': 1, 'center_y': 0.5},
                            padding='20dp',
                            allow_selection=True,
                            allow_copy=True,
                        ),
                    ),
                )
                modal_box.add_widget(MDFloatLayout(
                        MDLabel(
                            text='Device ID: ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                        MDButton(
                            MDButtonText(text='Show Device ID'),
                            style='filled',
                            pos_hint={'right': 1, 'center_y': .5},
                            on_release=lambda *args: dialog_info('device_id', state, view),
                            padding='30dp',
                        ),
                    ),
                )
                modal_box.add_widget(MDFloatLayout(
                        MDLabel(
                            text='API Key: ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                        MDButton(
                            MDButtonText(text='Show API Key'),
                            style='filled',
                            pos_hint={'right': 1, 'center_y': .5},
                            on_release=lambda *args: dialog_info('api_key', state, view),
                            padding='30dp',
                        ),
                    ),
                )
            modal_box.add_widget(MDFloatLayout(
                        MDLabel(
                            text='Anonymous Data Collection: ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                        MDCheckbox(
                            pos_hint={'right': 1, 'center_y': .5},
                            padding='40dp',
                            id='anon_data_'+device_id,
                            active=anon_data_coll,
                        ),
                        id='anon_data_float',
                    ),
            )
            modal_box.add_widget(MDFloatLayout(
                        MDLabel(
                            text='Firmware Version: ',
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.5},
                            padding='30dp',
                        ),
                        MDLabel(
                            text=state['state']['reported']['firmware_version'],
                            adaptive_size=True,
                            pos_hint={'right': 1, 'center_y': 0.5},
                            padding='30dp',
                            allow_selection=True,
                            allow_copy=True,
                        ),
                    ),
            )
            if 'admin' in state['state']:
                if show_advanced:
                    modal_box.height = str(modal_box_height + 550) + 'dp'
                    modal_box.add_widget(
                        MDFloatLayout(MDLabel(
                            text='Local Only Information',
                            adaptive_size=True,
                            font_style='Title',
                            role='large',
                            padding='5dp',
                            pos_hint={'center_x': 0.5, 'top': 1},
                        ))
                    )
                    modal_box.add_widget(MDFloatLayout(
                            MDLabel(
                                text='Debug: ',
                                adaptive_size=True,
                                pos_hint={'left': 1, 'center_y': 0.5},
                                padding='30dp',
                            ),
                            MDCheckbox(
                                pos_hint={'right': 1, 'center_y': .5},
                                padding='40dp',
                                id='debug_'+device_id,
                                active=state['state']['admin']['debug'],
                            ),
                            id='debug_float',
                        )
                    )
                    modal_box.add_widget(MDFloatLayout(
                            MDLabel(
                                text='Salt: ',
                                adaptive_size=True,
                                pos_hint={'left': 1, 'center_y': 0.5},
                                padding='30dp',
                            ),
                            MDButton(
                                MDButtonText(text='Show Salt'),
                                style='filled',
                                pos_hint={'right': 1, 'center_y': .5},
                                on_release=lambda *args: dialog_info('salt', state, view),
                                padding='30dp',
                            ),
                        )
                    )
                    modal_box.add_widget(
                        MDFloatLayout(MDLabel(
                            text='Danger Zone',
                            adaptive_size=True,
                            font_style='Title',
                            role='large',
                            padding='5dp',
                            pos_hint={'center_x': 0.5, 'top': 1},
                        ))
                    )
                    modal_box.add_widget(MDFloatLayout(
                            MDLabel(
                                text='Network Reset: ',
                                adaptive_size=True,
                                pos_hint={'left': 1, 'center_y': 0.5},
                                padding='30dp',
                            ),
                            MDCheckbox(
                                pos_hint={'right': 1, 'center_y': .5},
                                padding='40dp',
                                id='network_reset_'+device_id,
                                active=False,
                            ),
                            id='network_reset_float',
                        )
                    )
                    modal_box.add_widget(MDFloatLayout(
                            MDLabel(
                                text='Factory Reset: ',
                                adaptive_size=True,
                                pos_hint={'left': 1, 'center_y': 0.5},
                                padding='30dp',
                            ),
                            MDCheckbox(
                                pos_hint={'right': 1, 'center_y': .5},
                                padding='40dp',
                                id='factory_reset_'+device_id,
                                active=False,
                            ),
                            id='factory_reset_float',
                        )
                    )
                modal_box.add_widget(MDFloatLayout(MDButton(
                        MDButtonText(text='Submit'),
                        style="elevated",
                        pos_hint={"center_x": 0.5, "center_y": 0.5},
                        on_release=lambda *args: load_control_device(device_id, desired_state, modal_box, view),
                    )),
                )
            else:
                if show_advanced:
                    modal_box.height = str(modal_box_height + 320) + 'dp'
                    modal_box.add_widget(
                        MDFloatLayout(MDLabel(
                            text='Danger Zone',
                            adaptive_size=True,
                            font_style='Title',
                            role='large',
                            padding='5dp',
                            pos_hint={'center_x': 0.5, 'top': 1},
                        ))
                    )
                    modal_box.add_widget(MDFloatLayout(
                            MDLabel(
                                text='Delete device: ',
                                adaptive_size=True,
                                pos_hint={'left': 1, 'center_y': 0.5},
                                padding='30dp',
                            ),
                            MDCheckbox(
                                pos_hint={'right': 1, 'center_y': .5},
                                padding='40dp',
                                id='del_device_'+device_id,
                                active=False,
                            ),
                            id='del_device_float',
                        )
                    )
                modal_box.add_widget(
                    MDFloatLayout(MDButton(
                        MDButtonText(text='Submit'),
                        style="elevated",
                        pos_hint={"center_x": 0.5, "center_y": 0.5},
                        on_release=lambda *args: load_control_device(device_id, desired_state, modal_box, view),
                    ))
                )
            modal_scroll.add_widget(modal_box)
            view.add_widget(modal_scroll)
            view.open()

        update_modal(state)

    
    def shadow_to_card(self, shadow: dict):
        theme = self.theme_cls.theme_style
        # Conditions
        temp = ''
        if int(shadow['state']['reported']['sensors']['environment']['temperature']) != -9999:
            temp += str(round(shadow['state']['reported']['sensors']['environment']['temperature'], 2)) + u' \u00B0C'
            temp_icon = 'thermometer-check'
            temp_color = (94/255, 1., 110/255, 1.) if theme == 'Dark' else (31/255, 109/255, 38/255, 1.) #'green'
        if shadow['state']['reported']['alerts']['temperature_alert']:
            if 'Error' in shadow['state']['reported']['alerts']['temperature_alert']:
                temp += u' Error'
                temp_icon = 'alert-outline'
                temp_color = 'red'
            else:
                temp += ' ' + shadow['state']['reported']['alerts']['temperature_alert']
                temp_icon = 'thermometer-alert'
                temp_color = (230/255, 107/255, 25/255, 1.) if theme == 'Dark' else (161/255, 76/255, 20/255, 1.) #orange
        else:
            temp_icon = 'thermometer-check'
            temp_color = (94/255, 1., 110/255, 1.) if theme == 'Dark' else (31/255, 109/255, 38/255, 1.) #'green'
        humid = ''
        if int(shadow['state']['reported']['sensors']['environment']['humidity']) != -9999:
            humid += str(round(shadow['state']['reported']['sensors']['environment']['humidity'], 2)) + ' %'
            humid_icon = 'water-percent'
            humid_color = (94/255, 1., 110/255, 1.) if theme == 'Dark' else (31/255, 109/255, 38/255, 1.) #'green'
        if shadow['state']['reported']['alerts']['humidity_alert']:
            if 'Error' in shadow['state']['reported']['alerts']['humidity_alert']:
                humid += u' Error'
                humid_icon = 'alert-outline'
                humid_color = 'red'
            else:
                humid +=  ' ' + shadow['state']['reported']['alerts']['humidity_alert']
                humid_icon = 'water-percent-alert'
                humid_color = (230/255, 107/255, 25/255, 1.) if theme == 'Dark' else (161/255, 76/255, 20/255, 1.) #orange
        else:
            humid_icon = 'water-percent'
            humid_color = (94/255, 1., 110/255, 1.) if theme == 'Dark' else (31/255, 109/255, 38/255, 1.) #'green'
        water_level = ''
        if int(shadow['state']['reported']['sensors']['water_level']) != -9999:
            water_level += str(round(shadow['state']['reported']['sensors']['water_level'], 2)) + ' %'
            water_icon = 'water-check'
            water_color = (94/255, 1., 110/255, 1.) if theme == 'Dark' else (31/255, 109/255, 38/255, 1.) #'green'
        if shadow['state']['reported']['alerts']['water_level_alert']:
            if 'Error' in shadow['state']['reported']['alerts']['water_level_alert']:
                water_level += u' Error'
                water_icon = 'alert-outline'
                water_color = 'red'
            else:
                water_level +=  ' ' + shadow['state']['reported']['alerts']['water_level_alert']
                water_icon = 'water-alert'
                water_color = (230/255, 107/255, 25/255, 1.) if theme == 'Dark' else (161/255, 76/255, 20/255, 1.) #orange
        else:
            water_icon = 'water-check'
            water_color = (94/255, 1., 110/255, 1.) if theme == 'Dark' else (31/255, 109/255, 38/255, 1.) #'green'
        # Light
        if bool(shadow['state']['reported']['light']['status']):
            light_status = 'On'
            light_status_icon = 'lightbulb-on-30'
            light_status_color = 'cyan' if theme == 'Dark' else (18/255, 49/255, 124/255, 1.) #blue
        else: 
            light_status = 'Off' 
            light_status_icon = 'lightbulb-off-outline'
            light_status_color = (213/255, 139/255, 1., 1.) if theme == 'Dark' else (173/255, 66/255, 1, 1) #purple
        if shadow['state']['reported']['light']['on'] != shadow['state']['reported']['light']['off']:
            light_sched = 'Scheduled'
            light_sched_icon = 'clock-check-outline'
            light_sched_color = (94/255, 1., 110/255, 1.) if theme == 'Dark' else (31/255, 109/255, 38/255, 1.) #'green'
        else:
            light_sched = 'Not Scheduled'
            light_sched_icon = 'progress-clock'
            light_sched_color = (230/255, 107/255, 25/255, 1.) if theme == 'Dark' else (161/255, 76/255, 20/255, 1.) #orange
        if bool(shadow['state']['reported']['light']['override']):
            light_over = 'Overriden'
            light_over_icon = 'lightbulb-alert-outline'
            light_over_color = (230/255, 107/255, 25/255, 1.) if theme == 'Dark' else (161/255, 76/255, 20/255, 1.) #orange
        else:
            light_over = 'No Overrides'
            light_over_icon = light_status_icon
            light_over_color = (94/255, 1., 110/255, 1.) if theme == 'Dark' else (31/255, 109/255, 38/255, 1.) #'green'
        # Pump
        if bool(shadow['state']['reported']['pump']['status']):
            pump_status = 'On'
            pump_status_icon = 'pump'
            pump_status_color = 'cyan' if theme == 'Dark' else (18/255, 49/255, 124/255, 1.) #blue
        else: 
            pump_status = 'Off' 
            pump_status_icon = 'pump-off'
            pump_status_color =  (213/255, 139/255, 1., 1.) if theme == 'Dark' else (187/255, 66/255, 1, 1) #purple
        if bool(shadow['state']['reported']['pump']['powered']):
            pump_power = 'Scheduled'
            pump_power_icon = 'timer-check-outline'
            pump_power_color = (94/255, 1., 110/255, 1.) if theme == 'Dark' else (31/255, 109/255, 38/255, 1.) #'green'
        else:
            pump_power = 'Not Scheduled'
            pump_power_icon = 'timer-off-outline'
            pump_power_color = (230/255, 107/255, 25/255, 1.) if theme == 'Dark' else (161/255, 76/255, 20/255, 1.) #orange
        if bool(shadow['state']['reported']['pump']['drain']):
            drain = 'Draining'
            drain_icon = 'water-pump'
            drain_color = (230/255, 107/255, 25/255, 1.) if theme == 'Dark' else (161/255, 76/255, 20/255, 1.) #orange
        else:
            drain = 'Not Draining'
            drain_icon = 'water-pump-off'
            drain_color = (94/255, 1., 110/255, 1.) if theme == 'Dark' else (31/255, 109/255, 38/255, 1.) #'green'

        return temp, temp_icon, temp_color, humid, humid_icon, humid_color, water_level, water_icon, water_color, light_status, light_status_icon, light_status_color, light_sched, light_sched_icon, light_sched_color, light_over, light_over_icon, light_over_color, pump_status, pump_status_icon, pump_status_color, pump_power, pump_power_icon, pump_power_color, drain, drain_icon, drain_color


    def build_state_card(self, device_id: str, connectivity: str):
        if 'local' in connectivity.lower():
            icon = 'router-network-wireless'
            try:
                state = loads(decrypt(self.local_store.get(device_id)['value'], key))
            except:
                state = loads(decrypt(self.shadow_store.get(device_id)['value'], key))
        else:
            icon = 'access-point'
            state = loads(decrypt(self.shadow_store.get(device_id)['value'], key))
        
        temp, temp_icon, temp_color, humid, humid_icon, humid_color, water_level, water_icon, water_color, light_status, light_status_icon, light_status_color, light_sched, light_sched_icon, light_sched_color, light_over, light_over_icon, light_over_color, pump_status, pump_status_icon, pump_status_color, pump_power, pump_power_icon, pump_power_color, drain, drain_icon, drain_color = self.shadow_to_card(state)
        card_content = MDBoxLayout(
                    MDFloatLayout(
                        MDLabel(
                            text=state['state']['reported']['friendly_name'] if 'friendly_name' in state['state']['reported'] else state['state']['reported']['serial'],
                            adaptive_size=True,
                            font_style='Headline',
                            role='small',
                            pos_hint={'center_x': 0.5, 'center_y': 0.5},
                        ),
                        MDIconButton(
                            icon=icon,
                            pos_hint={'right': 1, 'center_y': 0.5},
                        ),
                        pos_hint={'center_x': 0.5, 'center_y': 0.5},
                    ),
                    MDBoxLayout(
                        MDFloatLayout(MDLabel(
                            text='Conditions',
                            adaptive_size=True,
                            font_style='Title',
                            role='large',
                            pos_hint={'center_x': 0.5, 'center_y': 0.5},
                        )),
                        MDBoxLayout(
                            height="15dp",
                            size_hint_y=None,
                        ),
                        MDBoxLayout(
                            MDIconButton(
                                icon='thermometer-lines',
                                adaptive_size=True,
                                pos_hint={'left': 1, 'center_y': 0.5},
                                font_size='10sp'
                            ),
                            MDLabel(
                                text='Temperature: ',
                                adaptive_size=False,
                                pos_hint={'left': 1, 'center_y': 0.5},
                            ),
                            MDIconButton(
                                icon=temp_icon,
                                adaptive_size=True,
                                pos_hint={'right': 1, 'center_y': 0.5},
                                font_size='10sp',
                                theme_icon_color="Custom",
                                icon_color=temp_color,
                            ),
                            MDLabel(
                                text=temp,
                                adaptive_size=True,
                                pos_hint={'right': 1, 'center_y': 0.5},
                                text_color=temp_color,
                            ),
                            orientation='horizontal'
                        ),
                        MDBoxLayout(
                            MDIconButton(
                                icon='water-percent',
                                adaptive_size=True,
                                pos_hint={'left': 1, 'center_y': 0.5},
                                font_size='10sp'
                            ),
                            MDLabel(
                                text='Humidity: ',
                                adaptive_size=False,
                                pos_hint={'left': 1, 'center_y': 0.5},
                            ),
                            MDIconButton(
                                icon=humid_icon,
                                adaptive_size=True,
                                pos_hint={'right': 1, 'center_y': 0.5},
                                font_size='10sp',
                                theme_icon_color="Custom",
                                icon_color=humid_color,
                            ),
                            MDLabel(
                                text=humid,
                                adaptive_size=True,
                                pos_hint={'right': 1, 'center_y': 0.5},
                                text_color=humid_color,
                            ),
                            orientation='horizontal'
                        ),
                        MDBoxLayout(
                            MDIconButton(
                                icon='water',
                                adaptive_size=True,
                                pos_hint={'left': 1, 'center_y': 0.5},
                                font_size='10sp'
                            ),
                            MDLabel(
                                text='Water Level: ',
                                adaptive_size=False,
                                pos_hint={'left': 1, 'center_y': 0.5},
                            ),
                            MDIconButton(
                                icon=water_icon,
                                adaptive_size=True,
                                pos_hint={'right': 1, 'center_y': 0.5},
                                font_size='10sp',
                                theme_icon_color="Custom",
                                icon_color=water_color,
                            ),
                            MDLabel(
                                text=water_level,
                                adaptive_size=True,
                                pos_hint={'right': 1, 'center_y': 0.5},
                                text_color=water_color,
                            ),
                            orientation='horizontal'
                        ),
                        orientation='vertical',
                        pos_hint={'center_x': 0.5, 'center_y': 0.5},
                    ),
                    MDBoxLayout(
                        MDFloatLayout(MDLabel(
                            text='Light',
                            adaptive_size=True,
                            font_style='Title',
                            role='large',
                            pos_hint={'center_x': 0.5, 'center_y': 0.5},
                        )),
                        MDBoxLayout(
                            MDIconButton(
                                icon=light_status_icon,
                                adaptive_size=True,
                                pos_hint={'left': 1, 'center_y': 0.5},
                                font_size='10sp',
                                #theme_icon_color="Custom",
                                #icon_color=light_status_color,
                            ),
                            MDLabel(
                                text=light_status,
                                adaptive_size=True,
                                pos_hint={'center_y': 0.5},
                                text_color=light_status_color,
                            ),
                            MDIconButton(
                                icon=light_sched_icon,
                                adaptive_size=True,
                                pos_hint={'left': 1, 'center_y': 0.5},
                                font_size='10sp',
                                #theme_icon_color="Custom",
                                #icon_color=light_sched_color,
                            ),
                            MDLabel(
                                text=light_sched,
                                adaptive_size=True,
                                pos_hint={'center_y': 0.5},
                                text_color=light_sched_color,
                            ),
                            MDIconButton(
                                icon=light_over_icon,
                                adaptive_size=True,
                                pos_hint={'left': 1, 'center_y': 0.5},
                                font_size='10sp',
                                #theme_icon_color="Custom",
                                #icon_color=light_over_color,
                            ),
                            MDLabel(
                                text=light_over,
                                adaptive_size=True,
                                pos_hint={'center_y': 0.5},
                                text_color=light_over_color,
                            ),
                            orientation='horizontal',
                        ),
                        orientation='vertical',
                        pos_hint={'center_x': 0.5, 'center_y': 0.5},
                    ),
                    MDBoxLayout(
                        MDFloatLayout(MDLabel(
                            text='Pump',
                            font_style='Title',
                            role='large',
                            adaptive_size=True,
                            pos_hint={'center_x': 0.5, 'center_y': 0.5},
                        )),
                        MDBoxLayout(
                            MDIconButton(
                                icon=pump_status_icon,
                                adaptive_size=True,
                                pos_hint={'left': 1, 'center_y': 0.5},
                                font_size='10sp',
                                #theme_icon_color="Custom",
                                #icon_color=pump_status_color,
                            ),
                            MDLabel(
                                text=pump_status,
                                adaptive_size=True,
                                pos_hint={'center_y': 0.5},
                                text_color=pump_status_color,
                            ),
                            MDIconButton(
                                icon=pump_power_icon,
                                adaptive_size=True,
                                pos_hint={'left': 1, 'center_y': 0.5},
                                font_size='10sp',
                                #theme_icon_color="Custom",
                                #icon_color=pump_power_color,
                            ),
                            MDLabel(
                                text=pump_power,
                                adaptive_size=True,
                                pos_hint={'center_y': 0.5},
                                text_color=pump_power_color,
                            ),
                            MDIconButton(
                                icon=drain_icon,
                                adaptive_size=True,
                                pos_hint={'left': 1, 'center_y': 0.5},
                                font_size='10sp',
                                #theme_icon_color="Custom",
                                #icon_color=drain_color,
                            ),
                            MDLabel(
                                text=drain,
                                adaptive_size=True,
                                pos_hint={'center_y': 0.5},
                                text_color=drain_color,
                            ),
                            orientation='horizontal',
                        ),
                        orientation='vertical',
                        pos_hint={'center_x': 0.5, 'center_y': 0.5},
                    ),
                    orientation='vertical',
                    padding='7dp',
                    halign='center',
                    spacing='5dp',
                    pos_hint={'center_x': 0.5, 'center_y': 0.5},
                )
        card_wid = MDCard(
                card_content,
                id='state_card_'+device_id,
                style='elevated', 
                padding='7dp', 
                ripple_behavior=True,
                size_hint=(None, None),
                size=("370dp", "400dp"),
                spacing='5dp',
                on_press=lambda *args: self.control_modal(state['state']['reported']['device_id'])
            )
        return card_wid


    def window_div_builder(self):
        try:
            state_card_list = []
            total_avail = self.locally_available+self.remotely_available
            if total_avail:
                state_card_list.append(MDBoxLayout(width="10dp", size_hint_x=None))
                for i in total_avail:
                    if i in self.locally_available:
                        state_card_list.append(self.build_state_card(i, 'Local Connection'))
                    else:
                        state_card_list.append(self.build_state_card(i, 'Remote Connection'))
                state_card_list.append(MDBoxLayout(width="10dp", size_hint_x=None))
                num_cards_window = Window.size[0]//390
                total_cards = len(total_avail)
                
                if num_cards_window:
                    if total_cards >= num_cards_window:
                        state_scroll_width_hint = 1 - (Window.size[0] - num_cards_window*390)/Window.size[0]
                    else:
                        state_scroll_width_hint = total_cards*390/Window.size[0]
                else:
                    state_scroll_width_hint = 1

            if self.locally_available_errored or self.local_only_unavailable or self.remote_unavail:
                errors_view = MDBoxLayout(
                        MDLabel(
                            text='Errors',
                            pos_hint={"center_x": 0.5},
                            adaptive_size=True,
                            font_style='Headline',
                            role='small',
                        ),
                        pos_hint={"center_x": 0.5, "center_y": 0.5},
                        orientation='vertical',
                    )
                if self.remote_unavail:
                    for i in self.remote_unavail:
                        try:
                            state = loads(decrypt(self.local_store.get(i)['value'], key))
                        except:
                            state = loads(decrypt(self.shadow_store.get(i)['value'], key))
                        errors_view.add_widget(
                            MDLabel(
                                text=state['state']['reported']['friendly_name'] if 'friendly_name' in state['state']['reported'] else state['state']['reported']['serial'],
                                pos_hint={"center_x": 0.5},
                                adaptive_size=True,
                                font_style='Title',
                                role='large',
                            )
                        )
                        errors_view.add_widget(
                            MDLabel(
                                text='Network connectivity issues!',
                                pos_hint={"center_x": 0.5},
                                adaptive_size=True,
                            )
                        )
                if self.locally_available_errored:
                    for i in self.locally_available_errored:
                        try:
                            state = loads(decrypt(self.local_store.get(i)['value'], key))
                        except:
                            state = loads(decrypt(self.shadow_store.get(i)['value'], key))
                        errors_view.add_widget(
                            MDLabel(
                                text=state['state']['reported']['friendly_name'] if 'friendly_name' in state['state']['reported'] else state['state']['reported']['serial'],
                                pos_hint={"center_x": 0.5},
                                adaptive_size=True,
                                font_style='Title',
                                role='large',
                            )
                        )
                        errors_view.add_widget(
                            MDLabel(
                                text='Connection errored! Please check stored API key!',
                                pos_hint={"center_x": 0.5},
                                adaptive_size=True,
                            )
                        )
                if self.local_only_unavailable:
                    for i in self.local_only_unavailable:
                        try:
                            state = loads(decrypt(self.local_store.get(i)['value'], key))
                        except:
                            state = loads(decrypt(self.shadow_store.get(i)['value'], key))
                        errors_view.add_widget(
                            MDLabel(
                                text=state['state']['reported']['friendly_name'] if 'friendly_name' in state['state']['reported'] else state['state']['reported']['serial'],
                                pos_hint={"center_x": 0.5},
                                adaptive_size=True,
                                font_style='Title',
                                role='large',
                            )
                        )
                        errors_view.add_widget(
                            MDLabel(
                                text='Local device outside of current network!',
                                pos_hint={"center_x": 0.5},
                                adaptive_size=True,
                            )
                        )
                errors_view_height = str(45*(len(self.locally_available_errored)+len(self.local_only_unavailable)+len(self.remote_unavail))+20)+'dp'
            else:
                errors_view = MDBoxLayout(
                        pos_hint={"center_x": 0.5, "center_y": 0.5},
                        orientation='vertical',
                    )
                errors_view_height = '1dp'
            print('Refreshed cards view')
            return state_card_list, state_scroll_width_hint, errors_view, errors_view_height
        except Exception as e:
            print('FAILED to refresh cards view')
            print(e)
            return [], 1, '', '100dp'
    

    def remote_control(self, shadow, device_id):
        user_c = {}
        for i in self.user_store.keys():
            user_c[i] = decrypt(self.user_store.get(i)['value'], key)
        topic = '$aws/things/' + user_c['thingName'] + '/shadow/name/' + device_id + '/update'
        topic = quote(topic)
        publish_url = 'https://iotcore.amastech.cloud:8443/topics/' + topic + '?qos=1'
        if not os.path.exists(data_path+'.tmp'):
            os.makedirs(data_path+'.tmp')
        if not os.path.exists(data_path+'.tmp/'+'mqtt_client_certs'):
            os.makedirs(data_path+'.tmp/'+'mqtt_client_certs')
        with open(data_path+'.tmp/'+'mqtt_client_certs'+'/cert.pem', 'w') as cert_file:
            cert_file.write(user_c['cert'])
            certfile = data_path+'.tmp/'+'mqtt_client_certs'+'/cert.pem'
        with open(data_path+'.tmp/'+'mqtt_client_certs'+'/key.pem', 'w') as key_file:
            key_file.write(user_c['key'])
            keyfile = data_path+'.tmp/'+'mqtt_client_certs'+'/key.pem'
        publish_msg = dumps(shadow)
        try:
            resp = requests.post(publish_url, data=publish_msg, cert=(certfile, keyfile), verify=certifi.where())
            shutil.rmtree(data_path+'.tmp/'+'mqtt_client_certs', ignore_errors=True)
        except Exception as e:
            print(e)


    def sub_client_on_connect(self, client, userdata, flags, reason_code, properties=None):
        user_c = {}
        for i in self.user_store.keys():
            user_c[i] = decrypt(self.user_store.get(i)['value'], key)

        print('Connected to IoT Thing: '+user_c['thingName'])
        # Build list of topics
        topics_sub = [('$aws/things/' + user_c['thingName'] + '/shadow/name/+/update/accepted',1)]
        client.subscribe(topics_sub)


    def sub_client_on_disconnect(self, client, userdata, flags, reason_code, properties=None):
        print('MQTT broker disconnected!')
        client.loop_stop()
        self.sub_client_reconnect()


    def sub_client_reconnect(self):
        sleep(600)
        print('Reconnecting to MQTT broker...')
        user_c = {}
        for i in self.user_store.keys():
            user_c[i] = decrypt(self.user_store.get(i)['value'], key)
        if not os.path.exists(data_path+'.tmp'):
            os.makedirs(data_path+'.tmp')
        if not os.path.exists(data_path+'.tmp/'+'mqtt_client_certs'):
            os.makedirs(data_path+'.tmp/'+'mqtt_client_certs')
        with open(data_path+'.tmp/'+'mqtt_client_certs'+'/cert.pem', 'w') as cert_file:
            cert_file.write(user_c['cert'])
        with open(data_path+'.tmp/'+'mqtt_client_certs'+'/key.pem', 'w') as key_file:
            key_file.write(user_c['key'])
        self.mqtt_sub.loop_start()
        try:
            sleep(30)
            shutil.rmtree(data_path+'.tmp/'+'mqtt_client_certs', ignore_errors=True)
        except:
            pass

        
    def sub_client_on_message(self, client, userdata, message):
        raw_payload = str(message.payload.decode("utf-8"))
        print(f"Received a message on topic: '{message.topic}', payload: '{raw_payload}'")
        payload = loads(raw_payload)
        if 'state' in payload and 'reported' in payload['state']:
            device_id = message.topic.split('/')[5]
            payload['state']['reported']['device_id'] = device_id
            self.save_remote_data(payload)

    
    def format_certs(self, string: str):
        if ('-----BEGIN CERTIFICATE-----' in string) and ('\n' not in string):
            string = string.split('-----BEGIN CERTIFICATE-----')[1].split('-----END CERTIFICATE-----')[0]
            string = '\n'.join(string[i:i+64] for i in range(0, len(string), 64))
            string = '-----BEGIN CERTIFICATE-----\n'+string+'\n-----END CERTIFICATE-----\n'
        elif ('-----BEGIN RSA PRIVATE KEY-----' in string) and ('\n' not in string):
            string = string.split('-----BEGIN RSA PRIVATE KEY-----')[1].split('-----END RSA PRIVATE KEY-----')[0]
            string = '\n'.join(string[i:i+64] for i in range(0, len(string), 64))
            string = '-----BEGIN RSA PRIVATE KEY-----\n'+string+'\n-----END RSA PRIVATE KEY-----\n'
        return string


    def get_remote_data(self, first_run=False):
        while not current_thread().stopped():
            if len(self.remotely_available) and self._internet.is_set():
                print('Starting guarded MQTT connection with helper function...')
                try:
                    user_c = {}
                    for i in self.user_store.keys():
                        user_c[i] = self.format_certs(decrypt(self.user_store.get(i)['value'], key))

                    if not self.mqtt_sub:
                        mqtt_host = 'iotcore.amastech.cloud'
                        print('Setting up MQTT connection for: '+user_c['thingName'])
                        self.mqtt_sub = mqtt.Client(mqtt.CallbackAPIVersion.VERSION1, client_id=user_c['mqtt_client_id'], protocol=mqtt.MQTTv5)

                        self.mqtt_sub.on_message = self.sub_client_on_message
                        self.mqtt_sub.on_connect = self.sub_client_on_connect
                        self.mqtt_sub.on_disconnect = self.sub_client_on_disconnect

                        if not os.path.exists(data_path+'.tmp'):
                            os.makedirs(data_path+'.tmp')
                        if not os.path.exists(data_path+'.tmp/'+'mqtt_client_certs'):
                            os.makedirs(data_path+'.tmp/'+'mqtt_client_certs')
                        with open(data_path+'.tmp/'+'mqtt_client_certs'+'/cert.pem', 'w') as cert_file:
                            cert_file.write(user_c['cert'])
                        with open(data_path+'.tmp/'+'mqtt_client_certs'+'/key.pem', 'w') as key_file:
                            key_file.write(user_c['key'])

                        self.mqtt_sub.tls_set(certfile=data_path+'.tmp/'+'mqtt_client_certs'+'/cert.pem',
                                        keyfile=data_path+'.tmp/'+'mqtt_client_certs'+'/key.pem',
                                        cert_reqs=ssl.CERT_REQUIRED,
                                        tls_version=ssl.PROTOCOL_TLSv1_2,
                                        ca_certs=certifi.where(),
                                        ciphers=None)

                        # Connect to sub client
                        properties_connect = Properties(PacketTypes.CONNECT)
                        properties_connect.SessionExpiryInterval = 30

                    if current_thread().stopped():
                        break

                    if not self.mqtt_sub.is_connected():
                        if not first_run and not current_thread().stopped():
                            sleep(600)
                        if current_thread().stopped():
                            break
                        self.mqtt_sub.connect(mqtt_host, 8883, clean_start=True, properties=properties_connect)
                        self.mqtt_sub.loop_start()
                        print('Connecting to AWS IoT Core...')

                    try:
                        sleep(30)
                        shutil.rmtree(data_path+'.tmp/'+'mqtt_client_certs', ignore_errors=True)
                    except:
                        pass

                    while not current_thread().stopped():
                        if not self.remotely_available:
                            self.mqtt_sub.loop_stop()
                            sleep(2)
                            break
                        else:
                            sleep(2)
                except Exception as e:
                    print(e)
                    sleep(2)
            else:
                try:
                    self.mqtt_sub.loop_stop()
                except:
                    pass
                sleep(2)
        

    def local_control(self, desired_shadow, device_id):
        try:
            response = requests.post('https://'+self.local_ips[device_id]['ip']+'/control', headers={'x-api-key': self.api_keys[device_id], 'Content-Type': 'application/json', 'Accept': 'application/json'}, json=desired_shadow, verify=base_dir+'assets/static/AMAS_CA.pem')
            if response:
                print('Got response!')
                self.local_store.put(device_id, value=encrypt(dumps(response.json()).encode(), key))
                print('Saved response!')
            else:
                print('Error controlling: '+device_id)
        except Exception as e:
            print(e)
            print('Error connecting to: '+device_id)


    def get_user_mqtt_details(self, token):
        try:
            cert_resp = requests.post('https://usercert.amastech.cloud/', json={'token': token}, headers={'Content-Type': 'application/json', 'Accept': 'application/json'}, verify=certifi.where())
        except: 
            print('Could not save user MQTT details')
            return {}
        user_mqtt_details = {}
        if cert_resp:
            cert_resp = cert_resp.json()
            for i in cert_resp:
                if i == 'cognito_id':
                    user_mqtt_details.update({'thingName': cert_resp[i]})
                else:
                    user_mqtt_details.update({i: cert_resp[i]})
        else:
            print('Did not save user MQTT details!')
        return user_mqtt_details


    def get_local_ips(self, shadow):
        local_ips = {}
        if shadow:
            for i in shadow:
                if 'internal_ip' in shadow[i]['state']['reported']:
                    if 'dhcp_hostname' in shadow[i]['state']['reported']:
                        local_ips.update({i: {'mdns': shadow[i]['state']['reported']['dhcp_hostname'], 'ip': shadow[i]['state']['reported']['internal_ip']}})
                    else:
                        local_ips.update({i: {'mdns': shadow[i]['state']['reported']['serial'], 'ip': shadow[i]['state']['reported']['internal_ip']}})
        return local_ips


    def get_api_keys(self, shadow):
        api_keys = {}
        if shadow:
            for i in shadow:
                if 'api_key' in shadow[i]['state']['reported']:
                    api_keys.update({i: shadow[i]['state']['reported']['api_key']})
        return api_keys


    def get_availability_lists(self, only_once=True, first_run=False):
        print('Creating availability list')
        if self.local_ips:
            for i in self.local_ips:
                self.local_store = JsonStore(data_path+'locals.json')
                try:
                    print('Testing: '+i)
                    response= requests.get('https://'+self.local_ips[i]['ip']+'/local', timeout=20, verify=base_dir+'assets/static/AMAS_CA.pem')
                    if response:
                        print(i+' is local')
                        if i not in self.locally_available:
                            self.locally_available.append(i)
                        if i in self.locally_available_errored:
                            self.locally_available_errored.pop(self.locally_available_errored.index(i))
                        if i in self.local_only_unavailable:
                            self.local_only_unavailable.pop(self.local_only_unavailable.index(i))
                        if i in self.remotely_available:
                            self.remotely_available.pop(self.remotely_available.index(i))
                    else:
                        print(i+' is local but errored')
                        if i not in self.locally_available_errored:
                            self.locally_available_errored.append(i)
                        if i in self.locally_available:
                            self.locally_available.pop(self.locally_available.index(i))
                        if i in self.local_only_unavailable:
                            self.local_only_unavailable.pop(self.local_only_unavailable.index(i))
                        if i in self.remotely_available:
                            self.remotely_available.pop(self.remotely_available.index(i))
                except Exception as e:
                    print(e)
                    if i in self.local_only:
                        print(i+' is local only, but outside the network')
                        if i not in self.local_only_unavailable:
                            self.local_only_unavailable.append(i)
                        if i in self.locally_available:
                            self.locally_available.pop(self.locally_available.index(i))
                        if i in self.locally_available_errored:
                            self.locally_available_errored.pop(self.locally_available_errored.index(i))
                        if i in self.remotely_available:
                            self.remotely_available.pop(self.remotely_available.index(i))
                        
                    else:
                        print(i+' is remote')
                        if i not in self.remotely_available:
                            self.remotely_available.append(i)
                        if i in self.locally_available:
                            self.locally_available.pop(self.locally_available.index(i))
                        if i in self.local_only_unavailable:
                            self.local_only_unavailable.pop(self.local_only_unavailable.index(i))
                        if i in self.locally_available_errored:
                            self.locally_available_errored.pop(self.locally_available_errored.index(i))
                        if i in self.local_store.keys():
                            self.local_store.delete(i)
            
            self.locally_available.sort()
            self.locally_available_errored.sort()
            self.local_only_unavailable.sort()
            self.remotely_available.sort()

            self.remote_unavail = []
            if not self._internet.is_set():
                for i in self.remotely_available:
                    self.remote_unavail.append(i)
                self.remotely_available = []

            print('Created availability list')
            print('locally available:')
            print(self.locally_available)
            print('locally available errored:')
            print(self.locally_available_errored)
            print('local only unavailable:')
            print(self.local_only_unavailable)
            print('remotely available:')
            print(self.remotely_available)
            print('remotely unavailable:')
            print(self.remote_unavail)


    def save_remote_data(self, data):
        self.shadow_store = JsonStore(data_path+'shadows.json')
        shadow_store = {}
        for i in self.shadow_store.keys():
            shadow_store[i] = loads(decrypt(self.shadow_store.get(i)['value'], key))
        try:
            print('Saving remote data for: '+data['state']['reported']['device_id'])
        except Exception as e:
            print(e)
        print(data)
        if 'state' in data and 'reported' in data['state']:
            if data['state']['reported']['device_id'] in shadow_store:
                self.load_diff(shadow_store[data['state']['reported']['device_id']]['state']['reported'],data['state']['reported'])
            else:
                shadow_store[data['state']['reported']['device_id']] = {}
                shadow_store[data['state']['reported']['device_id']]['state'] = {} 
                shadow_store[data['state']['reported']['device_id']]['state']['reported'] = data['state']['reported']
            print('Saved remote data for: '+data['state']['reported']['device_id'])
        if 'state' in data and 'metadata' in data['state']:
            if data['state']['reported']['device_id'] in shadow_store:
                self.load_diff(shadow_store[data['state']['reported']['device_id']]['state']['metadata'],data['state']['metadata'])
            else:
                shadow_store[data['state']['reported']['device_id']] = {}
                shadow_store[data['state']['reported']['device_id']]['state'] = {} 
                shadow_store[data['state']['reported']['device_id']]['state']['metadata'] = data['state']['metadata']
            print('Saved remote metadata for: '+data['state']['reported']['device_id'])

        self.shadow_store.put(data['state']['reported']['device_id'], value=encrypt(dumps(shadow_store[data['state']['reported']['device_id']]).encode(), key))
        print('Synced with encrypted shadow storage!')
        try:
            self.refresh_single_card(data['state']['reported']['device_id'])
        except Exception as e:
            print(e)


    def save_local_data(self, data):
        self.local_store = JsonStore(data_path+'locals.json')
        local_store = {}
        for i in self.local_store.keys():
            local_store[i] = loads(decrypt(self.local_store.get(i)['value'], key))
        print(data)
        if 'state' in data and 'reported' in data['state']:
            local_store[data['state']['reported']['device_id']] = data
            self.local_store.put(data['state']['reported']['device_id'], value=encrypt(dumps(data).encode(), key), timestamp=time.time())
            print('Saved local data for: '+data['state']['reported']['device_id'])
            self.refresh_single_card(data['state']['reported']['device_id'])

    
    def get_local_data_poll(self):
        try:
            while True:
                self.local_store = JsonStore(data_path+'locals.json')
                if self.locally_available:
                    for i in self.locally_available:
                        try:
                            print('Connecting to: '+i)
                            resp = requests.get('https://'+self.local_ips[i]['ip']+'/poll', headers={'x-api-key': self.api_keys[i], 'Content-Type': 'application/json', 'Accept': 'application/json'}, verify=base_dir+'assets/static/AMAS_CA.pem', timeout=31)
                            if resp:
                                print('Got local data for: '+i)
                                self.save_local_data(resp.json())
                            else:
                                print('Errored while getting local data for: '+i)
                        except:
                            print('Failed to get local data for: '+i)
                sleep(60)
        except:
            self.get_local_data_poll()  

    
    async def threaded_ws_connections(self, url, device_id, ssl_context, stop):
        async with aiohttp.ClientSession() as session:
            self.local_ws_connections[device_id] = session
            async with session.ws_connect(url=url,ssl_context=ssl_context) as ws:
                await ws.send_json({'x-api-key': self.api_keys[device_id]})
                print('Sent API key for authentication to: '+device_id)
                async for msg in ws:
                    if msg.type == aiohttp.WSMsgType.TEXT:
                        if msg.data == 'close cmd':
                            del self.local_ws_connections[device_id]
                            await ws.close()
                            break
                        else:
                            print('Got local data from: '+device_id)
                            self.save_local_data(loads(msg.data))
                    elif msg.type == aiohttp.WSMsgType.ERROR:
                        del self.local_ws_connections[device_id]
                        await ws.close()
                        break


    async def await_aiohttp_loop(self):
        counter = 0
        while not current_thread().stopped():
            if counter < 30:
                counter += 1
            else:
                try:
                    counter = 0
                    if len(self.local_ws_connections) != len(self.locally_available):
                        raise Exception
                    for i in self.local_ws_connections:
                        if not i in self.locally_available:
                            raise Exception
                except:
                    for i in self.local_ws_connections:
                        try:
                            await self.local_ws_connections[i].close()
                        except Exception as e:
                            print(e)
            await asyncio.sleep(2)
        for i in self.local_ws_connections:
            try:
                self._local_data_loop.create_task(self.local_ws_connections[i].close())
            except Exception as e:
                print(e)


    def get_local_data_ws(self):
        self._local_data_loop = asyncio.get_event_loop_policy().new_event_loop()
        asyncio.set_event_loop(self._local_data_loop)
        try:
            if self.locally_available:
                ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
                ssl_context.load_verify_locations(base_dir+'assets/static/AMAS_CA.pem')
                print('Creating local WebSocket connections')
                self.local_ws_connections = {}
                stop_threads = False
                for i in self.locally_available:
                    print('Opening WebSocket connection to: '+i)
                    url = 'wss://'+self.local_ips[i]['ip']+'/metrics?dev_id='+i
                    print('Connecting to: '+url)
                    self.local_ws_connections[i] = self._local_data_loop.create_task(self.threaded_ws_connections(url, i, ssl_context, lambda: stop_threads))
                self._local_data_loop.run_until_complete(self.await_aiohttp_loop())
        except Exception as e:
            print(e)
            stop_threads = True
            if not current_thread().stopped():
                sleep(60)
                self.get_local_data_ws()


    def no_devices_window(self):
        new_device_label = MDLabel(
                        text='If you just set up a new device with your user account, it may take some time to show up, please wait!',
                        #font_style='Title',
                        #role='large',
                        pos_hint={"center_x": 0.5, "top": 1},
                        adaptive_size=True,
                        padding='30dp',
                    )
        #new_device_label.bind(width=lambda *x: new_device_label.setter('text_size')(new_device_label, (new_device_label.width, None),texture_size=lambda *x: new_device_label.setter('height')(new_device_label, new_device_label.texture_size[1])))
        local_device_label = MDLabel(
                        text='If you just set up a new local only device, you will have to manually add the existing local device.',
                        #font_style='Title',
                        #role='large',
                        pos_hint={"center_x": 0.5, "top": 0.8},
                        adaptive_size=True,
                        padding='30dp',
                    )
        #local_device_label.bind(width=lambda *x: local_device_label.setter('text_size')(local_device_label, (local_device_label.width, None),texture_size=lambda *x: local_device_label.setter('height')(local_device_label, local_device_label.texture_size[1])))
        card = MDFloatLayout(
                    new_device_label,
                    local_device_label,
                    MDButton(
                        MDButtonText(text='Setup a new device'),
                        style="elevated",
                        pos_hint={"center_x": 0.5, "top": 0.5},
                        on_release=lambda *args: self.redirect_screen('setup'),
                    ),
                    MDButton(
                        MDButtonText(text='Add existing local device'),
                        style="elevated",
                        pos_hint={"center_x": 0.5, "top": 0.3},
                        on_release=lambda *args: self.redirect_screen('local_add'),
                    ),
                    size_hint = (0.8, 0.8)
                )
        return card


    def load_logout_del_device(self):
        self.show_loading()
        Thread(target=self.logout_del_device, daemon=True, name='logout_del_device').start()

    def logout_del_device(self):
        try:
            self.mqtt_sub.loop_stop()
        except Exception as e:
            print(e)
        try:
            self._update_check_thread.stop()
            while self._update_check_thread.is_alive():
                try:
                    self._update_check_thread.stop()
                    sleep(0.1)
                except:
                    print('Stopped update check')
                    break
        except Exception as e:
            print(e)
        try:
            self._remote_data_thread.stop()
            while self._remote_data_thread.is_alive():
                try:
                    self._remote_data_thread.stop()
                    sleep(0.1)
                except:
                    print('Stopped remote data collection')
                    break
        except Exception as e:
            print(e)
        try:
            self._local_data_thread.stop()
            while self._local_data_thread.is_alive():
                try:
                    self._local_data_thread.stop()
                    sleep(0.1)
                except:
                    print('Stopped local data collection')
                    break
        except Exception as e:
            print(e)
        try:
            self._local_data_loop.stop()
        except Exception as e:
            print(e)
        try:
            self._new_device_discovery_thread.stop()
            while self._new_device_discovery_thread.is_alive():
                try:
                    self._new_device_discovery_thread.stop()
                    sleep(0.1)
                except:
                    print('Stopped new device discovery service')
                    break
        except Exception as e:
            print(e)
        try:
            self._coordinator_thread.stop()
            while self._coordinator_thread.is_alive():
                try:
                    self._coordinator_thread.stop()
                    sleep(0.1)
                except:
                    print('Stopped task co-ordinator')
                    break
        except Exception as e:
            print(e)

        try:
            os.remove(data_path+'shadows.json')
        except: pass
        try:
            os.remove(data_path+'user.json')
        except:pass
        try:
            os.remove(data_path+'cookies.json')
        except:pass
        try:
            os.remove(data_path+'locals.json')
        except:pass

        self.shadow_store = JsonStore(data_path+'shadows.json')
        self.cookie_jar = JsonStore(data_path+'cookies.json')
        self.user_store = JsonStore(data_path+'user.json')
        self.local_store = JsonStore(data_path+'locals.json')

        self.cookie_jar.put('wait_for_validation', value=True)

        self.account_menu.dismiss()
        self.dismiss_loading()

        self.redirect_screen('logout_screen')

    def load_logout_only(self):
        self.show_loading()
        Thread(target=self.logout_only, daemon=True, name='logout_only').start()

    def logout_only(self):
        try:
            self.mqtt_sub.loop_stop()
        except Exception as e:
            print(e)
        try:
            self._update_check_thread.stop()
            while self._update_check_thread.is_alive():
                try:
                    self._update_check_thread.stop()
                    sleep(0.1)
                except:
                    print('Stopped update check')
                    break
        except Exception as e:
            print(e)
        try:
            self._remote_data_thread.stop()
            while self._remote_data_thread.is_alive():
                try:
                    self._remote_data_thread.stop()
                    sleep(0.1)
                except:
                    print('Stopped remote data collection')
                    break
        except Exception as e:
            print(e)
        try:
            self._local_data_thread.stop()
            while self._local_data_thread.is_alive():
                try:
                    self._local_data_thread.stop()
                    sleep(0.1)
                except:
                    print('Stopped local data collection')
                    break
        except Exception as e:
            print(e)
        try:
            self._local_data_loop.stop()
        except Exception as e:
            print(e)
        try:
            self._new_device_discovery_thread.stop()
            while self._new_device_discovery_thread.is_alive():
                try:
                    self._new_device_discovery_thread.stop()
                    sleep(0.1)
                except:
                    print('Stopped new device discovery service')
                    break
        except Exception as e:
            print(e)
        try:
            self._coordinator_thread.stop()
            while self._coordinator_thread.is_alive():
                try:
                    self._coordinator_thread.stop()
                    sleep(0.1)
                except:
                    print('Stopped task co-ordinator')
                    break
        except Exception as e:
            print(e)

        for i in self.shadow_store.keys():
            if i in self.local_store.keys():
                self.local_store.delete(i)
        try:
            os.remove(data_path+'shadows.json')
        except: pass
        try:
            os.remove(data_path+'user.json')
        except:pass
        try:
            os.remove(data_path+'cookies.json')
        except:pass

        self.shadow_store = JsonStore(data_path+'shadows.json')
        self.cookie_jar = JsonStore(data_path+'cookies.json')
        self.user_store = JsonStore(data_path+'user.json')
        
        self.cookie_jar.put('local_only', value=True)

        self.account_menu.dismiss()
        self.dismiss_loading()

        self.redirect_screen('logout_screen')


    def new_device_discovery(self):
        while not current_thread().stopped():
            try:
                local_store = JsonStore(data_path+'locals.json')
                shadow_store = JsonStore(data_path+'shadows.json')

                if (local_store.count() != self.local_store.count()) or (shadow_store.count() != len(self.locally_available) + len(self.locally_available_errored) + len(self.local_only_unavailable) + len(self.remotely_available) + len(self.remote_unavail) - len(self.local_only)):
                    print('New device found!')
                    self.local_store = local_store
                    self.shadow_store = shadow_store
                    self.initialize()
                    self.get_availability_lists()
                    self.refresh_entire_window()
            except Exception as e:
                print(e)

            sleep(2)


    def initialize(self, shadows={}, user_t={}, only_once=True, first_run=False):
        # Check shadow recheck
        if shadows and user_t:
            if 'recheck' in shadows and 'refresh_token' in user_t:
                # Get shadows
                try:
                    del shadows['recheck']
                    print('Getting new shadows...')
                    if self._internet.is_set():
                        resp = requests.get('https://amastech.cloud/shadows', headers={'Content-Type': 'application/json'}, json={'refresh_token': user_t['refresh_token'], 'identity_token': user_t['identity_token'], 'timestamp': user_t['timestamp'].replace('"', '')}, verify=certifi.where())
                    else:
                        raise ConnectionError
                    if resp:
                        print('Got new shadows!')
                        shadows = resp.json()
                        if shadows:
                            for i in shadows:
                                if i in ['identity_token', 'timestamp']:
                                    self.cookie_jar.put(i, value=encrypt(str(shadows[i]).encode(), key))
                                else:
                                    self.shadow_store.put(i, value=encrypt(dumps(shadows[i]).encode(), key))
                            for i in ['identity_token', 'timestamp']:
                                if i in shadows:
                                    del shadows[i]
                            print('Saved new shadows in encrypted shadow storage')
                except Exception as e:
                    print(e)
        try:
            if 'recheck' in self.shadow_store.keys():
                self.shadow_store.delete('recheck')    
            # Update remote shadow store
            if shadows and ('recheck' not in shadows):
                if self.shadow_store.count():
                    for i in self.shadow_store.keys():
                        if i not in shadows:
                            self.shadow_store.delete(i)
                for i in shadows:
                    self.shadow_store.put(i, value=encrypt(dumps(shadows[i]).encode(), key))
                print('Synced shadow store')
            else:
                if self.shadow_store.count():
                    for i in self.shadow_store.keys():
                        if not i == 'recheck':
                            shadows[i] = loads(decrypt(self.shadow_store.get(i)['value'], key))

            if 'recheck' in shadows:
                del shadows['recheck']
            # Get local IPs from remote shadow store
            print('Getting local IPs from shadow store')
            self.local_ips = self.get_local_ips(shadows)

            # Get device tokens from remote shadow store
            print('Getting device tokens from shadow store')
            self.api_keys = self.get_api_keys(shadows)

            # Get a list of local only devices if applicable
            self.local_only = []
            if self.local_store.count():
                for i in self.local_store.keys():
                    if not i in self.shadow_store.keys():
                        self.local_only.append(i)

            # Create master local IP and device token lists (including local only devices)
            local_states = {}
            for i in self.local_store.keys():
                local_states[i] = loads(decrypt(self.local_store.get(i)['value'], key))
            
            if self.local_only:
                print('Local only devices known')
                for i in self.local_only:
                    self.local_ips[i] = self.get_local_ips({i: local_states[i]})[i]
                    self.api_keys[i] = self.get_api_keys({i: local_states[i]})[i]
            else:
                print('No local only devices known')
            
            # Show cached data on first run
            if first_run:
                if shadows:
                    for i in shadows:
                        self.remotely_available.append(i)
                if self.local_only:
                    for i in self.local_only:
                        self.locally_available.append(i)
                Clock.schedule_once(self.refresh_entire_window)

            # Check if user has MQTT certs in IndexedDB if refresh token is found
            if user_t:
                if 'refresh_token' in user_t:
                    if self.user_store.count() == 4 and 'cert' in self.user_store.keys() and 'key' in self.user_store.keys() and 'mqtt_client_id' in self.user_store.keys() and 'thingName' in self.user_store.keys():
                        print('User store has all required data')
                    else:
                        print('Getting user MQTT details')
                        user_mqtt_details = self.get_user_mqtt_details(user_t['identity_token'])
                        if user_mqtt_details:
                            for i in user_mqtt_details:
                                self.user_store.put(i, value=encrypt(str(user_mqtt_details[i]).encode(), key))
                            print('Synced user store')
                        else: print("Couldn't get user MQTT details")
        except Exception as e:
            print(e)


    @mainthread
    def refresh_entire_window(self, only_once=True, first_run=False, cache_run=False):
        try:
            # Edit active window
            print('Editing active window')
            if len(self.locally_available) + len(self.locally_available_errored) + len(self.local_only_unavailable) + len(self.remotely_available) == 0:
                print('No known devices found!')
                no_state_card = self.no_devices_window()
                self.dismiss_loading()
                self.ids.state_card_view.clear_widgets()
                self.ids.dash_anchor.size_hint_y = None
                self.ids.dash_anchor.height = '1dp'
                self.ids.dash_inner_scroll.size_hint_y = None
                self.ids.dash_inner_scroll.height = '1dp'
                self.ids.state_card_view.size_hint_y = None
                self.ids.state_card_view.height = '1dp'
                self.ids.errors_view.clear_widgets()
                self.ids.errors_view.height = '400dp'
                self.ids.errors_view.add_widget(no_state_card)
            else:
                self.ids.state_card_view.clear_widgets()
                self.ids.errors_view.clear_widgets()
                self.ids.dash_anchor.height = '1dp'
                self.ids.dash_inner_scroll.height = '1dp'
                self.ids.state_card_view.height = '1dp'
                state_cards, state_scroll_width_hint, errors, errors_view_height = self.window_div_builder()
                if state_cards:
                    self.ids.dash_anchor.height = '400dp'
                    self.ids.dash_inner_scroll.height = '400dp'
                    self.ids.state_card_view.height = '400dp'
                    self.ids.dash_inner_scroll.size_hint_x = state_scroll_width_hint
                    for i in state_cards:
                        self.ids.state_card_view.add_widget(i)
                if errors:
                    self.ids.errors_view.height = errors_view_height
                    self.ids.errors_view.add_widget(errors)
        except Exception as e:
            print(e)
    

    @mainthread
    def refresh_single_card(self, device_id):
        try:
            if device_id in self.locally_available:
                state_card = self.build_state_card(device_id, 'Local Connection')
            elif device_id in self.remotely_available:
                state_card = self.build_state_card(device_id, 'Remote Connection')
            else:
                return False
            for i in self.ids['state_card_view'].children:
                if i.id:
                    if device_id in i.id:
                        card_content = state_card.children[0]
                        state_card.clear_widgets()
                        i.clear_widgets()
                        i.add_widget(card_content)
                        print('Refreshed state card for: '+device_id)
        except Exception as e:
            print(e)


    @mainthread
    def update_avail(self, extension):
        if ARCH:
            arch = '/' + ARCH
        else:
            arch = ''
        if MODE == 'portable':
            filename = '/AMASDash%20'+PLATFORM.capitalize()+'%20Portable'+extension
        else:
            filename = '/AMASTech%20Dashboard'+extension
        self.ids.update_view.clear_widgets()
        self.ids.update_view.add_widget(
            MDButton(
                MDButtonText(text='Update available! Please download new version!'),
                style="text",
                pos_hint={"center_x": 0.5, "center_y": 0.5},
                on_release=lambda *args: webbrowser.open('https://amastech.cloud/client/'+PLATFORM+arch+filename),
            )
        )


    def first_run_tasks(self):
        global wait_for_resize_refresh

        while True:
            cookie_jar = JsonStore(data_path+'cookies.json')
            if cookie_jar.count() and 'wait_for_validation' not in cookie_jar.keys():
                break
            else:
                if self.manager:
                    if self.manager.current == 'dashboard':
                        self.show_loading()
                print('Waiting for cookies to be baked!')
                sleep(2)

        self.show_loading()
        self.refresh_menu_items()

        self.cookie_jar = cookie_jar
        shadows = {'recheck': True}
        user_t = {}
        if self.cookie_jar.count():
            for i in self.cookie_jar.keys():
                if i not in ['wait_for_validation', 'local_only']:
                    user_t[i] = decrypt(self.cookie_jar.get(i)['value'], key)
        self.initialize(shadows, user_t, first_run=True)
        self.dismiss_loading()
        self.get_availability_lists(first_run=True)
        self.refresh_entire_window(first_run=True)

        wait_for_resize_refresh = False
        
        self._coordinator_thread = StoppableThread(target=self.task_coordinator, daemon=True, name='task-coordinator')
        self._coordinator_thread.start()
    

    def check_update(self):
        counter = 0
        while not current_thread().stopped():
            if not counter:
                if 'version' in self.settings_store.keys():
                    local_version = self.settings_store.get('version')['value']
                    print('Installed version: '+local_version)
                    if PLATFORM == 'linux':
                        if MODE == 'portable':
                            extension = '.tar.xz'
                        elif MODE == 'appimage':
                            extension = '.AppImage'
                        elif MODE == 'deb':
                            extension = '.deb'
                        elif MODE == 'rpm':
                            extension = '.rpm'
                        else:
                            extension = '.flatpak'
                    elif PLATFORM == 'windows':
                        if MODE == 'portable':
                            extension = '.zip'
                        else:
                            extension = '.exe'
                    elif PLATFORM == 'mac':
                        extension = '.dmg'
                    else:
                        extension = False
                    if extension:
                        if ARCH:
                            resp = requests.get('https://amastech.cloud/client/'+PLATFORM+'/'+ARCH+'/version', verify=certifi.where())
                        else:
                            resp = requests.get('https://amastech.cloud/client/'+PLATFORM+'/version', verify=certifi.where())
                        if resp:
                            resp.encoding = 'UTF-8'
                            remote_version = resp.text.strip()
                            print('Upsteam version found: '+remote_version)
                            remote_version = remote_version.split('.')
                            local_version = local_version.split('.')
                            if remote_version != local_version:
                                if int(local_version[0].split('v')[1]) < int(remote_version[0].split('v')[1]):
                                    self.update_avail(extension)
                                elif int(local_version[0].split('v')[1]) > int(remote_version[0].split('v')[1]):
                                    print('No update necessary')
                                elif int(local_version[1]) < int(remote_version[1]):
                                    self.update_avail(extension)
                                elif int(local_version[1]) > int(remote_version[1]):
                                    print('No update necessary')
                                elif int(local_version[2]) >= int(remote_version[2]):
                                    print('No update necessary')
                                else:
                                    self.update_avail(extension)
                            else:
                                print('No update necessary')
                        else:
                            print('Could not get upstream version')
                    else:
                        print('Update can only be done from the app store')
            if counter < 7200:
                counter += 1
            else:
                counter = 0
            sleep(2)

    def task_coordinator(self):
        print('Started dashboard task coordinator...')
        self._local_data_thread = StoppableThread(target=self.get_local_data_ws, daemon=True, name='local-data-ws')
        self._local_data_thread.start()
        self._new_device_discovery_thread = StoppableThread(target=self.new_device_discovery, daemon=True, name='newly-added-device')
        self._new_device_discovery_thread.start()
        self._update_check_thread = StoppableThread(target=self.check_update, daemon=True, name='update-check')
        self._update_check_thread.start()
        if not 'local_only' in self.cookie_jar.keys():
            self._remote_data_thread = StoppableThread(target=self.get_remote_data, args=[True], daemon=True, name='MQTT-helper')
            self._remote_data_thread.start()

        counter = 0
        while not current_thread().stopped():
            sleep(2)
            while True:
                cookie_jar = JsonStore(data_path+'cookies.json')
                if cookie_jar.count() and 'wait_for_validation' not in cookie_jar.keys():
                    self.cookie_jar = cookie_jar
                    break
                else:
                    print('Waiting for cookies to be baked!')
                    sleep(2)
            user_t = {}
            if self.cookie_jar.count():
                for i in self.cookie_jar.keys():
                    if i != 'local_only':
                        user_t[i] = decrypt(self.cookie_jar.get(i)['value'], key)
                    else:
                        user_t[i] = True
            if counter < 36000:
                counter += 1
                if not counter%300:
                    self.get_availability_lists()
                    self.refresh_entire_window()
            else:
                counter = 0
                if not 'local_only' in user_t:
                    req_resp = requests.post(url='https://amastech.cloud/authorize',headers={'Content-Type': 'application/json'}, json={'refresh_token': user_t['refresh_token'], 'identity_token': user_t['identity_token'], 'timestamp': user_t['timestamp']}, verify=certifi.where())
                    if req_resp:
                        updated_tokens = req_resp.json()
                        for i in updated_tokens:
                            self.cookie_jar.put(i, value=encrypt(str(updated_tokens[i]).encode(), key))
                            cookie_jar.put(i, value=encrypt(str(updated_tokens[i]).encode(), key))
                        self.initialize(shadows={'recheck': True}, user_t=user_t)
                        self.get_availability_lists()
                        self.refresh_entire_window()