 # coding: utf-8

from kivy.storage.jsonstore import JsonStore
from kivymd.uix.screen import MDScreen
from kivy.clock import Clock
from kivy.uix.camera import Camera
from kivy.uix.modalview import ModalView
from kivy.core.window import Window
from kivy.uix.widget import Widget
from kivy.clock import mainthread
from kivymd.uix.menu import MDDropdownMenu
from kivymd.uix.divider import MDDivider
from kivymd.uix.label import MDLabel
from kivymd.uix.progressindicator import MDCircularProgressIndicator
from kivymd.uix.dialog import (
    MDDialog,
    MDDialogIcon,
    MDDialogHeadlineText,
    MDDialogSupportingText,
    MDDialogButtonContainer,
    MDDialogContentContainer,
)
from kivymd.uix.boxlayout import MDBoxLayout
from kivymd.uix.floatlayout import MDFloatLayout
from kivymd.uix.button import MDButton, MDButtonText
from kivymd.uix.snackbar import (
    MDSnackbar,
    MDSnackbarCloseButton,
    MDSnackbarSupportingText,
    MDSnackbarButtonContainer
)
from json import dumps, loads
import requests, time, cv2, os


from assets.env import *


import warnings
warnings.showwarning = print


class LocalAddScreen(MDScreen):
    def __init__(self, **kwargs):
        super(LocalAddScreen, self).__init__(**kwargs)

        self.loading_dialog = ModalView(
            auto_dismiss=False,
            size_hint=(1,1),
            background_color=(0,0,0,0)
        )
        self.loading_dialog.add_widget(
            MDFloatLayout(
                MDCircularProgressIndicator(
                    size_hint=(None,None),
                    size=('24dp','24dp'),
                    pos_hint={'center_x': 0.5, 'center_y': 0.5},
                )
            ),
        )

        self.hamburger_menu_items = [
            {
                "text": 'Dashboard',
                "on_release": lambda *args: self.redirect_screen('dashboard'),
            },
            {
                "text": 'Setup a new device',
                "on_release": lambda *args: self.redirect_screen('setup'),
            },
            {
                "text": 'Add existing local device',
                "on_release": lambda *args: self.redirect_screen('local_add'),
            },
            {
                "text": 'Manual Update',
                "on_release": lambda *args: self.redirect_screen('update'),
            },
            {
                "text": 'Switch Dark/Light Mode',
                "on_release": lambda *args: self.change_display_mode(),
            },
            {
                "text": 'About',
                "on_release": lambda *args: self.redirect_screen('about'),
            },
        ]

        self.qr_management_menu_items = [
            {
                "text": 'Load from saved QR codes',
                "on_release": lambda *args: self.saved_qrs_modal(),
            },
            {
                "text": 'Scan new QR code',
                "on_release": lambda *args: self.new_qr(),
            },
        ]
        
        self.settings_store = JsonStore(data_path+'settings.json')
        self.local_store = JsonStore(data_path+'locals.json')
        self.shadow_store = JsonStore(data_path+'shadows.json')
        self.cookie_jar = JsonStore(data_path+'cookies.json')
        self.user_store = JsonStore(data_path+'user.json')
        self.qr_store = JsonStore(data_path+'qrs.json')
        self.all_qrs = {}
        self.qr_dialog = MDDialog()

        self.hamburger_menu = MDDropdownMenu(
            caller=self.ids.hamburger_button,
            items=self.hamburger_menu_items,
        )

        self.qr_management_menu = MDDropdownMenu(
            caller=self.ids.qr_button,
            items=self.qr_management_menu_items,
        )

    
    @mainthread
    def show_loading(self):
        try:
            self.loading_dialog.open()
        except Exception as e:
            print(e)


    @mainthread
    def dismiss_loading(self):
        try:
            self.loading_dialog.dismiss()
        except Exception as e:
            print(e)


    def saved_qrs_modal(self):
        self.qr_dialog = MDDialog()
        self.qr_store = JsonStore(data_path+'qrs.json')

        # Create master list and a dialog modal
        self.all_qrs = {}
        if self.qr_store.count():
            for i in self.qr_store.keys():
                self.all_qrs[i] = decrypt(self.qr_store.get(i)['value'], key)

        dialog_content = MDDialogContentContainer(orientation="vertical")
        dialog_content.add_widget(MDDivider())
        
        if self.all_qrs:
            for i in self.all_qrs:
                dialog_content.add_widget(
                    MDButton(
                        MDButtonText(text=i),
                        style="text",
                        id=i+'_qr',
                        on_release=self.load_saved_qr,
                        pos_hint={'center_x': 0.5}
                    )
                )
        else:
            dialog_content.add_widget(
                MDLabel(
                    text='No saved QR codes!',
                    adaptive_size=True,
                    padding='8dp',
                    pos_hint={'center_x': 0.5}
                )
            )
        dialog_content.add_widget(MDDivider())

        self.qr_dialog.add_widget(
            MDDialogIcon(
                icon='backup-restore'
            ))
        self.qr_dialog.add_widget(MDDialogHeadlineText(
                text='Load from saved QR codes'
            ))
        self.qr_dialog.add_widget(MDDialogSupportingText(
                text='Load device serial number and API key from a list of previously saved QR codes.'
            ))
        self.qr_dialog.add_widget(dialog_content)
        self.qr_dialog.add_widget(MDDialogButtonContainer(
                Widget(),
                MDButton(
                    MDButtonText(text="Cancel"),
                    style="text",
                    on_release=lambda *args: self.dismiss_qr_dialog(self.qr_dialog), 
                )
            )
        )
        
        self.qr_dialog.open()

    
    def dismiss_qr_dialog(self, dialog: MDDialog):
        dialog.dismiss()

    
    def load_saved_qr(self, *args, **kwargs):
        serial = args[0].id.split('_')[0]
        self.ids.serial_txt.text = serial
        self.ids.api_txt.text = self.all_qrs[serial]
        self.qr_dialog.dismiss()
        self.qr_management_menu.dismiss()

    
    def stop_camera(self, cam: Camera, view: ModalView):
        cam.play = False
        view.dismiss()
        print('Camera stopped')


    def take_photo(self, cam: Camera, view: ModalView):
        temper=str(time.time())
        cam.export_to_png(data_path+'.tmp/temp_'+temper+'.png')
        image = cv2.imread(data_path+'.tmp/temp_'+temper+'.png')
        qrCodeDetector = cv2.QRCodeDetector()
        try:
            decodedText, points, _ = qrCodeDetector.detectAndDecode(image)
            data = loads(decodedText)
            print(data)
            os.remove(data_path+'.tmp/temp_'+temper+'.png')
            self.qr_store = JsonStore(data_path+'qrs.json')
            self.qr_store.put(data['serial'], value=encrypt(data['api_key'].encode(), key))
            self.ids.serial_txt.text = data['serial']
            self.ids.api_txt.text = data['api_key']
            cam.play = False
            view.dismiss()
            self.qr_management_menu.dismiss()
            print('QR code found!')
        except:
            os.remove(data_path+'.tmp/temp_'+temper+'.png')
            MDSnackbar(
                MDSnackbarSupportingText(
                    text='No QR code detected!'
                ),
                MDSnackbarButtonContainer(
                    MDSnackbarCloseButton(
                        icon='close'
                    ),
                    pos_hint={"center_y": 0.5}
                ),
                y='24dp',
                orientation="horizontal",
                pos_hint={"center_x": 0.5},
                size_hint_x=0.5,
            ).open()
            print('QR code not detected!')


    def new_qr(self):
        try:
            width_hint = self.settings_store.get('modal_width_hint')['value']
        except:
            try:
                width = Window.size[0]
                if width < 600:
                    self.settings_store.put('modal_width_hint', value=0.9)
                    width_hint = 0.9
                elif width < 800:
                    self.settings_store.put('modal_width_hint', value=0.8)
                    width_hint = 0.8
                else:
                    self.settings_store.put('modal_width_hint', value=0.7)
                    width_hint = 0.7
            except:
                width_hint = 0.8
        cam = Camera(
                resolution=(400,400),
            )
        if width_hint:
            view = ModalView(size_hint=(width_hint, 0.7), border=(80, 80, 80, 80), auto_dismiss=False,)
        else:
            view = ModalView(size_hint=(0.8, 0.7), border=(80, 80, 80, 80), auto_dismiss=False,)
        modal_box = MDBoxLayout(
            padding='25dp',
            md_bg_color=self.theme_cls.backgroundColor,
            orientation='vertical',
        )
        modal_box.add_widget(cam)
        modal_box.add_widget(MDFloatLayout(
            MDButton(
                MDButtonText(
                    text='Dismiss'
                ),
                pos_hint={"center_x": 0.3, "center_y": 0.5},
                on_release=lambda *args: self.stop_camera(cam, view),
            ),
            MDButton(
                MDButtonText(
                    text='Capture'
                ),
                pos_hint={"center_x": 0.7, "center_y": 0.5},
                on_release=lambda *args: self.take_photo(cam, view),
            ), 
            size_hint_y=None, 
            height='50dp'
        ))
        view.add_widget(modal_box)
        view.open()
        cam.play = True


    def load_local_add(self):
        self.show_loading()
        Thread(target=self.local_add, daemon=True, name='add-local-device').start()


    def local_add(self):
        serial = self.ids.serial_txt.text
        print('Got serial: '+serial)
        api_key = self.ids.api_txt.text
        print('Got API key too')
        ip = self.ids.ip_txt.text

        print('Looking for AMAS Tower Device at: '+ip)
        try:
            resp = requests.get('https://' + ip + '/poll', headers={'x-api-key': api_key, 'Content-Type': 'application/json', 'Accept': 'application/json'}, verify=base_dir+'assets/static/AMAS_CA.pem')
            if resp:
                try:
                    state = resp.json()
                    device_id = state['state']['reported']['device_id']
                    print('Found AMAS Tower Device at: '+ip)
                    self.local_store.put(device_id, value=encrypt(dumps(state).encode(), key))
                    print('Saved device in local encrypted storage')
                    Clock.schedule_once(self.redirect_screen)
                except:
                    print('Could not connect to AMAS Tower Device at: '+ip)        
            elif resp.status_code == 401:
                print('Found AMAS Tower Device at: '+ip+', but could not authenticate!')
            else:
                print('There is a device at: '+ip+', but could not verify if it is a AMAS Tower Device!')
        except:
            print('Could not connect to AMAS Tower Device at: '+ip)
        self.dismiss_loading()


    def change_display_mode(self):
        self.theme_cls.theme_style = (
            "Dark" if self.theme_cls.theme_style == "Light" else "Light"
        )
        self.settings_store.put('theme', value=self.theme_cls.theme_style)


    def redirect_screen(self, screen = 'dashboard'):
        self.hamburger_menu.dismiss()
        self.dismiss_loading()
        self.manager.current = screen
