 # coding: utf-8

from kivy.storage.jsonstore import JsonStore
from kivy.uix.widget import Widget
from kivy.uix.modalview import ModalView
from kivy.clock import mainthread
from kivymd.uix.screen import MDScreen
from kivymd.uix.menu import MDDropdownMenu
from kivymd.uix.divider import MDDivider
from kivymd.uix.label import MDLabel
from kivymd.uix.floatlayout import MDFloatLayout
from kivymd.uix.progressindicator import MDCircularProgressIndicator
from kivymd.uix.dialog import (
    MDDialog,
    MDDialogIcon,
    MDDialogHeadlineText,
    MDDialogSupportingText,
    MDDialogButtonContainer,
    MDDialogContentContainer,
)
from kivymd.uix.button import MDButton, MDButtonText 
from json import loads
import webbrowser, asyncio, ssl
import aiohttp

from assets.env import *
from assets.python.Console import Console


import warnings
warnings.showwarning = print


class UpdateScreen(MDScreen):
    def __init__(self, **kwargs):
        super(UpdateScreen, self).__init__(**kwargs)

        self.hamburger_menu_items = [
            {
                "text": 'Dashboard',
                "on_release": lambda *args: self.redirect_screen('dashboard'),
            },
            {
                "text": 'Setup a new device',
                "on_release": lambda *args: self.redirect_screen('setup'),
            },
            {
                "text": 'Add existing local device',
                "on_release": lambda *args: self.redirect_screen('local_add'),
            },
            {
                "text": 'Manual Update',
                "on_release": lambda *args: self.redirect_screen('update'),
            },
            {
                "text": 'Switch Dark/Light Mode',
                "on_release": lambda *args: self.change_display_mode(),
            },
            {
                "text": 'About',
                "on_release": lambda *args: self.redirect_screen('about'),
            },
        ]
        
        self.loading_dialog = ModalView(
            auto_dismiss=False,
            size_hint=(1,1),
            background_color=(0,0,0,0)
        )
        self.loading_dialog.add_widget(
            MDFloatLayout(
                MDCircularProgressIndicator(
                    size_hint=(None,None),
                    size=('24dp','24dp'),
                    pos_hint={'center_x': 0.5, 'center_y': 0.5},
                )
            ),
        )

        self.settings_store = JsonStore(data_path+'settings.json')
        self.local_store = JsonStore(data_path+'locals.json')
        self.shadow_store = JsonStore(data_path+'shadows.json')
        self.cookie_jar = JsonStore(data_path+'cookies.json')
        self.user_store = JsonStore(data_path+'user.json')

        self.hamburger_menu = MDDropdownMenu(
            caller=self.ids.hamburger_button,
            items=self.hamburger_menu_items,
        )

        logDir = data_path
        logFile = "update.log"
        self.console = Console(logDir,logFile)
        self.ids.update_console.clear_widgets()
        self.ids.update_console.add_widget(self.console)
        self.ids.update_console.md_bg_color = (1,1,1,1) if self.theme_cls.theme_style == "Light" else (0,0,0,1)
        self.console_text_color = 'gray'
        self.selected_device_id = None

        self.warning_dialog = MDDialog(
            MDDialogIcon(
                icon='alert-octagon'
            ),
            MDDialogHeadlineText(
                text='Manual updates are NOT recommended'
            ),
            MDDialogSupportingText(
                text='Due to the constrained resource limits of ESP32, we do NOT recommend manual updates.'
                ' The update process of the MicroPython firmware is quite involved, and may need several retries.'
                ' Due to the clearly advanced nature of the update process, we do not provide support for manual updates from within our app, and you\'d need to do rely on the community if any problem arises;'
                ' we strongly recommend that you read our API documentation to get familiar with the update process! \n\nBy clicking "I Agree", you agree that you are technologically knowledgable who can (and has) read out API documentation. '
            ),
            MDDialogButtonContainer(
                Widget(),
                MDButton(
                    MDButtonText(text="Go back to Dashboard"),
                    style="text",
                    on_release=lambda *args: self.redirect_screen('dashboard'),
                ),
                MDButton(
                    MDButtonText(text="API Docs"),
                    style="text",
                    on_release=lambda *args: webbrowser.open('https://api.amastech.cloud/'),
                ),
                MDButton(
                    MDButtonText(text="I Agree"),
                    style="text",
                    on_release=self.dismiss_warning,
                ),
            ),
            auto_dismiss=False,
        )

        self.device_choice_dialog = MDDialog()

        # Create master list and a dialog modal
        self.all_local_states = {}
        if self.local_store.count():
            for i in self.local_store.keys():
                self.all_local_states[i] = loads(decrypt(self.local_store.get(i)['value'], key))

        dialog_content = MDDialogContentContainer(orientation="vertical")
        dialog_content.add_widget(MDDivider())
        
        if self.all_local_states:
            for i in self.all_local_states:
                dialog_content.add_widget(
                    MDButton(
                        MDButtonText(text=self.all_local_states[i]['state']['reported']['serial'] + (' (' + self.all_local_states[i]['state']['reported']['friendly_name'] + ')') if 'friendly_name' in self.all_local_states[i]['state']['reported'] else ''),
                        style="text",
                        id=i+'_load',
                        on_release=self.load_saved_device,
                        pos_hint={'center_x': 0.5}
                    )
                )
        else:
            dialog_content.add_widget(
                MDLabel(
                    text='No saved devices!',
                    adaptive_size=True,
                    padding='8dp',
                    pos_hint={'center_x': 0.5}
                )
            )
        dialog_content.add_widget(MDDivider())

        self.device_choice_dialog.add_widget(
            MDDialogIcon(
                icon='backup-restore'
            ))
        self.device_choice_dialog.add_widget(MDDialogHeadlineText(
                text='Choose from local devices'
            ))
        self.device_choice_dialog.add_widget(MDDialogSupportingText(
                text='Choose the local device that you want to manually update'
            ))
        self.device_choice_dialog.add_widget(dialog_content)
        self.device_choice_dialog.add_widget(MDDialogButtonContainer(
                Widget(),
                MDButton(
                    MDButtonText(text="Cancel"),
                    style="text",
                    on_release=lambda *args: self.dismiss_device_dialog(self.device_choice_dialog), 
                )
            )
        )


    def on_enter(self, *args):
        try:
            self.warning_dialog.open()
        except:
            pass
        self.ids.update_console.md_bg_color = (1,1,1,1) if self.theme_cls.theme_style == "Light" else (0,0,0,1)
        self.local_store = JsonStore(data_path+'locals.json')
        self.all_local_states = {}
        if self.local_store.count():
            for i in self.local_store.keys():
                self.all_local_states[i] = loads(decrypt(self.local_store.get(i)['value'], key))


    @mainthread
    def show_loading(self):
        try:
            self.loading_dialog.open()
        except Exception as e:
            print(e)


    @mainthread
    def dismiss_loading(self):
        try:
            self.loading_dialog.dismiss()
        except Exception as e:
            print(e)


    def dismiss_warning(self, *args):
        try:
            self.warning_dialog.dismiss()
        except:
            pass


    def reset_selection(self, *args):
        try:
            self.dismiss_device_dialog()
        except:
            pass
        try:
            self.dismiss_warning()
        except:
            pass
        self.selected_device_id = None
        self.ids.device_chooser_txt.text = 'Choose device'
        self.console.clear()


    def load_saved_device(self, *args, **kwargs):
        device_id = args[0].id.split('_')[0]
        self.selected_device_id = device_id
        self.ids.device_chooser_txt.text = self.all_local_states[device_id]['state']['reported']['serial']
        self.device_choice_dialog.dismiss()


    def change_display_mode(self):
        self.theme_cls.theme_style = (
            "Dark" if self.theme_cls.theme_style == "Light" else "Light"
        )
        self.ids.update_console.md_bg_color = (1,1,1,1) if self.theme_cls.theme_style == "Light" else (0,0,0,1)
        self.settings_store.put('theme', value=self.theme_cls.theme_style)


    def dismiss_device_dialog(self, dialog: MDDialog):
        dialog.dismiss()


    @mainthread
    def console_note(self, note: str):
        self.console.note(note, self.console_text_color)


    def start_update(self, *args):
        self.show_loading()
        Thread(target=self._update, args=['start'], daemon=True, name='manual-device-update').start()


    def tail_update(self, *args):
        self.show_loading()
        Thread(target=self._update, args=['tail'], daemon=True, name='manual-device-update').start()


    def _update(self, update_dic):
        if self.selected_device_id:
            if update_dic == 'start':
                update_dic = {'start': True}
            else:
                update_dic = {'tail': True}
            _local_data_loop = asyncio.get_event_loop_policy().new_event_loop()
            asyncio.set_event_loop(_local_data_loop)
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            ssl_context.load_verify_locations(base_dir+'assets/static/AMAS_CA.pem')
            url = 'wss://'+self.all_local_states[self.selected_device_id]['state']['reported']['internal_ip']+'/update'
            _local_data_loop.run_until_complete(self.threaded_update(url, ssl_context, update_dic))


    async def threaded_update(self, url, ssl_context, update_dic):
        async with aiohttp.ClientSession() as session:
            async with session.ws_connect(url=url,ssl_context=ssl_context) as ws:
                await ws.send_json({'x-api-key': self.all_local_states[self.selected_device_id]['state']['reported']['api_key']})
                first_response = await ws.receive()
                print(first_response.data)
                if 'Previous run' in first_response.data:
                    print(first_response.data)
                    self.console_note(first_response.data)
                    async for msg in ws:
                        if msg.type == aiohttp.WSMsgType.TEXT:
                            if msg.data == 'close cmd':
                                await ws.close()
                                break
                            elif 'End previous run' in msg.data:
                                print(msg.data)
                                self.console_note(msg.data)
                                break
                            else:
                                print(msg.data)
                                self.console_note(msg.data)
                        elif msg.type == aiohttp.WSMsgType.ERROR:
                            await ws.close()
                            break
                else:
                    await asyncio.sleep(5)
                await ws.send_json(update_dic)
                self.dismiss_loading()
                async for msg in ws:
                    if msg.type == aiohttp.WSMsgType.TEXT:
                        if msg.data == 'close cmd':
                            await ws.close()
                            break
                        else:
                            print(msg.data)
                            self.console_note(msg.data)
                    elif msg.type == aiohttp.WSMsgType.ERROR:
                        await ws.close()
                        break
        self.dismiss_loading()


    def redirect_screen(self, screen = 'dashboard'):
        self.hamburger_menu.dismiss()
        try:
            self.warning_dialog.dismiss()
        except:
            pass
        self.manager.current = screen
