 # coding: utf-8

from kivymd.uix.screen import MDScreen
from kivymd.uix.menu import MDDropdownMenu
from kivymd.uix.button import MDButton, MDButtonText
from kivymd.uix.progressindicator import MDCircularProgressIndicator
from kivymd.uix.dialog import (
    MDDialog,
    MDDialogIcon,
    MDDialogHeadlineText,
    MDDialogSupportingText,
    MDDialogButtonContainer,
    MDDialogContentContainer,
)
from kivymd.uix.divider import MDDivider
from kivymd.uix.label import MDLabel
from kivymd.uix.boxlayout import MDBoxLayout
from kivymd.uix.floatlayout import MDFloatLayout
from kivymd.uix.button import MDButton, MDButtonText
from kivymd.uix.snackbar import (
    MDSnackbar,
    MDSnackbarCloseButton,
    MDSnackbarSupportingText,
    MDSnackbarButtonContainer
)
from kivy.storage.jsonstore import JsonStore
from kivy.uix.widget import Widget
from kivy.clock import mainthread
from kivy.uix.camera import Camera
from kivy.uix.modalview import ModalView
from kivy.core.window import Window
from json import loads
import requests, time, cv2, os, certifi

from assets.env import *


import warnings
warnings.showwarning = print


class SetupScreen(MDScreen):
    def __init__(self, **kwargs):
        super(SetupScreen, self).__init__(**kwargs)

        self.loading_dialog = ModalView(
            auto_dismiss=False,
            size_hint=(1,1),
            background_color=(0,0,0,0)
        )
        self.loading_dialog.add_widget(
            MDFloatLayout(
                MDCircularProgressIndicator(
                    size_hint=(None,None),
                    size=('24dp','24dp'),
                    pos_hint={'center_x': 0.5, 'center_y': 0.5},
                )
            ),
        )

        self.hamburger_menu_items = [
            {
                "text": 'Dashboard',
                "on_release": lambda *args: self.redirect_screen('dashboard'),
            },
            {
                "text": 'Setup a new device',
                "on_release": lambda *args: self.redirect_screen('setup'),
            },
            {
                "text": 'Add existing local device',
                "on_release": lambda *args: self.redirect_screen('local_add'),
            },
            {
                "text": 'Manual Update',
                "on_release": lambda *args: self.redirect_screen('update'),
            },
            {
                "text": 'Switch Dark/Light Mode',
                "on_release": lambda *args: self.change_display_mode(),
            },
            {
                "text": 'About',
                "on_release": lambda *args: self.redirect_screen('about'),
            },
        ]

        self.qr_management_menu_items = [
            {
                "text": 'Load from saved devices',
                "on_release": lambda *args: self.saved_devices_modal(),
            },
            {
                "text": 'Load from saved QR codes',
                "on_release": lambda *args: self.saved_qrs_modal(),
            },
            {
                "text": 'Scan new QR code',
                "on_release": lambda *args: self.new_qr(),
            },
        ]
        
        self.settings_store = JsonStore(data_path+'settings.json')
        self.local_store = JsonStore(data_path+'locals.json')
        self.shadow_store = JsonStore(data_path+'shadows.json')
        self.cookie_jar = JsonStore(data_path+'cookies.json')
        self.user_store = JsonStore(data_path+'user.json')
        self.qr_store = JsonStore(data_path+'qrs.json')
        self.all_states = {}
        self.all_qrs = {}
        self.device_dialog = MDDialog()
        self.qr_dialog = MDDialog()

        self.hamburger_menu = MDDropdownMenu(
            caller=self.ids.hamburger_button,
            items=self.hamburger_menu_items,
        )

        self.qr_management_menu = MDDropdownMenu(
            caller=self.ids.qr_button,
            items=self.qr_management_menu_items,
        )

    
    @mainthread
    def show_loading(self):
        try:
            self.loading_dialog.open()
        except Exception as e:
            print(e)


    @mainthread
    def dismiss_loading(self):
        try:
            self.loading_dialog.dismiss()
        except Exception as e:
            print(e)
    

    def saved_devices_modal(self):
        self.device_dialog = MDDialog()
        self.local_store = JsonStore(data_path+'locals.json')
        self.shadow_store = JsonStore(data_path+'shadows.json')

        # Get a list of local only devices if applicable
        local_only = []
        if self.local_store.count():
            for i in self.local_store.keys():
                if not i in self.shadow_store.keys():
                    local_only.append(i)

        # Create master list and a dialog modal
        self.all_states = {}
        if self.shadow_store.count():
            for i in self.shadow_store.keys():
                self.all_states[i] = loads(decrypt(self.shadow_store.get(i)['value'], key))
        if local_only:
            for i in local_only:
                self.all_states[i] = loads(decrypt(self.local_store.get(i)['value'], key))

        dialog_content = MDDialogContentContainer(orientation="vertical")
        dialog_content.add_widget(MDDivider())
        
        if self.all_states:
            for i in self.all_states:
                dialog_content.add_widget(
                    MDButton(
                        MDButtonText(text=self.all_states[i]['state']['reported']['serial'] + (' (' + self.all_states[i]['state']['reported']['friendly_name'] + ')') if 'friendly_name' in self.all_states[i]['state']['reported'] else ''),
                        style="text",
                        id=i+'_load',
                        on_release=self.load_saved_device,
                        pos_hint={'center_x': 0.5}
                    )
                )
        else:
            dialog_content.add_widget(
                MDLabel(
                    text='No saved devices!',
                    adaptive_size=True,
                    padding='8dp',
                    pos_hint={'center_x': 0.5}
                )
            )
        dialog_content.add_widget(MDDivider())

        self.device_dialog.add_widget(
            MDDialogIcon(
                icon='backup-restore'
            ))
        self.device_dialog.add_widget(MDDialogHeadlineText(
                text='Load from saved devices'
            ))
        self.device_dialog.add_widget(MDDialogSupportingText(
                text='Load device serial number and API key from a list of saved devices.'
            ))
        self.device_dialog.add_widget(dialog_content)
        self.device_dialog.add_widget(MDDialogButtonContainer(
                Widget(),
                MDButton(
                    MDButtonText(text="Cancel"),
                    style="text",
                    on_release=lambda *args: self.dismiss_device_dialog(self.device_dialog), 
                )
            )
        )
        
        self.device_dialog.open()

    
    def dismiss_device_dialog(self, dialog: MDDialog):
        dialog.dismiss()

    
    def load_saved_device(self, *args, **kwargs):
        device_id = args[0].id.split('_')[0]
        self.ids.serial_txt.text = self.all_states[device_id]['state']['reported']['serial']
        self.ids.api_txt.text = self.all_states[device_id]['state']['reported']['api_key']
        self.device_dialog.dismiss()
        self.qr_management_menu.dismiss()

    
    def saved_qrs_modal(self):
        self.qr_dialog = MDDialog()
        self.qr_store = JsonStore(data_path+'qrs.json')

        # Create master list and a dialog modal
        self.all_qrs = {}
        if self.qr_store.count():
            for i in self.qr_store.keys():
                self.all_qrs[i] = decrypt(self.qr_store.get(i)['value'], key)

        dialog_content = MDDialogContentContainer(orientation="vertical")
        dialog_content.add_widget(MDDivider())
        
        if self.all_qrs:
            for i in self.all_qrs:
                dialog_content.add_widget(
                    MDButton(
                        MDButtonText(text=i),
                        style="text",
                        id=i+'_qr',
                        on_release=self.load_saved_qr,
                        pos_hint={'center_x': 0.5}
                    )
                )
        else:
            dialog_content.add_widget(
                MDLabel(
                    text='No saved QR codes!',
                    adaptive_size=True,
                    padding='8dp',
                    pos_hint={'center_x': 0.5}
                )
            )
        dialog_content.add_widget(MDDivider())

        self.qr_dialog.add_widget(
            MDDialogIcon(
                icon='backup-restore'
            ))
        self.qr_dialog.add_widget(MDDialogHeadlineText(
                text='Load from saved QR codes'
            ))
        self.qr_dialog.add_widget(MDDialogSupportingText(
                text='Load device serial number and API key from a list of previously saved QR codes.'
            ))
        self.qr_dialog.add_widget(dialog_content)
        self.qr_dialog.add_widget(MDDialogButtonContainer(
                Widget(),
                MDButton(
                    MDButtonText(text="Cancel"),
                    style="text",
                    on_release=lambda *args: self.dismiss_qr_dialog(self.qr_dialog), 
                )
            )
        )
        
        self.qr_dialog.open()

    
    def dismiss_qr_dialog(self, dialog: MDDialog):
        dialog.dismiss()

    
    def load_saved_qr(self, *args, **kwargs):
        serial = args[0].id.split('_')[0]
        self.ids.serial_txt.text = serial
        self.ids.api_txt.text = self.all_qrs[serial]
        self.qr_dialog.dismiss()
        self.qr_management_menu.dismiss()

    
    def stop_camera(self, cam: Camera, view: ModalView):
        cam.play = False
        view.dismiss()
        print('Camera stopped')


    def take_photo(self, cam: Camera, view: ModalView):
        temper=str(time.time())
        cam.export_to_png(data_path+'.tmp/temp_'+temper+'.png')
        image = cv2.imread(data_path+'.tmp/temp_'+temper+'.png')
        qrCodeDetector = cv2.QRCodeDetector()
        try:
            decodedText, points, _ = qrCodeDetector.detectAndDecode(image)
            data = loads(decodedText)
            print(data)
            os.remove(data_path+'.tmp/temp_'+temper+'.png')
            self.qr_store = JsonStore(data_path+'qrs.json')
            self.qr_store.put(data['serial'], value=encrypt(data['api_key'].encode(), key))
            self.ids.serial_txt.text = data['serial']
            self.ids.api_txt.text = data['api_key']
            cam.play = False
            view.dismiss()
            self.qr_management_menu.dismiss()
            print('QR code found!')
        except:
            os.remove(data_path+'.tmp/temp_'+temper+'.png')
            MDSnackbar(
                MDSnackbarSupportingText(
                    text='No QR code detected!'
                ),
                MDSnackbarButtonContainer(
                    MDSnackbarCloseButton(
                        icon='close'
                    ),
                    pos_hint={"center_y": 0.5}
                ),
                y='24dp',
                orientation="horizontal",
                pos_hint={"center_x": 0.5},
                size_hint_x=0.5,
            ).open()
            print('QR code not detected!')


    def new_qr(self):
        try:
            width_hint = self.settings_store.get('modal_width_hint')['value']
        except:
            try:
                width = Window.size[0]
                if width < 600:
                    self.settings_store.put('modal_width_hint', value=0.9)
                    width_hint = 0.9
                elif width < 800:
                    self.settings_store.put('modal_width_hint', value=0.8)
                    width_hint = 0.8
                else:
                    self.settings_store.put('modal_width_hint', value=0.7)
                    width_hint = 0.7
            except:
                width_hint = 0.8
        cam = Camera(
                resolution=(400,400),
            )
        if width_hint:
            view = ModalView(size_hint=(width_hint, 0.7), border=(80, 80, 80, 80), auto_dismiss=False,)
        else:
            view = ModalView(size_hint=(0.8, 0.7), border=(80, 80, 80, 80), auto_dismiss=False,)
        modal_box = MDBoxLayout(
            padding='25dp',
            md_bg_color=self.theme_cls.backgroundColor,
            orientation='vertical',
        )
        modal_box.add_widget(cam)
        modal_box.add_widget(MDFloatLayout(
            MDButton(
                MDButtonText(
                    text='Dismiss'
                ),
                pos_hint={"center_x": 0.3, "center_y": 0.5},
                on_release=lambda *args: self.stop_camera(cam, view),
            ),
            MDButton(
                MDButtonText(
                    text='Capture'
                ),
                pos_hint={"center_x": 0.7, "center_y": 0.5},
                on_release=lambda *args: self.take_photo(cam, view),
            ), 
            size_hint_y=None, 
            height='50dp'
        ))
        view.add_widget(modal_box)
        view.open()
        cam.play = True


    def on_enter(self):
        if ('refresh_token' in self.cookie_jar.keys()) and ('local_only' not in self.cookie_jar.keys()):
            self.ids.registration_checkbox.state = 'down'

    
    def change_display_mode(self):
        self.theme_cls.theme_style = (
            "Dark" if self.theme_cls.theme_style == "Light" else "Light"
        )
        self.settings_store.put('theme', value=self.theme_cls.theme_style)


    def load_setup(self):
        self.show_loading()
        Thread(target=self.setup, daemon=True, name='validate-pre-setup-device').start()


    def setup(self):
        self.cookie_jar = JsonStore(data_path+'cookies.json')
        cookie_jar = JsonStore(data_path+'cookies.json')
        if cookie_jar.exists('refresh_token'):
            try:
                refresh_token = decrypt(cookie_jar.get('refresh_token')['value'], key)
                print('Found refresh token, looking for identity token...')
                if cookie_jar.exists('identity_token') and cookie_jar.exists('timestamp'):
                    try:
                        identity_token = decrypt(cookie_jar.get('identity_token')['value'], key)
                        timestamp = decrypt(cookie_jar.get('timestamp')['value'], key).replace('"', '')
                        print('Found identity token!')
                    except:
                        print('Identity token corrupted, app will try to get new ones using the refresh token!')
                        identity_token = None
                        timestamp = None
            except:
                print('Refresh token corrupted, please login!')
                refresh_token = None
                identity_token = None
                timestamp = None
        else:
            print('Refresh token not found.')
            refresh_token = None

        if refresh_token:
            print('Validating existing session...')
            try:
                req_resp = requests.post(url='https://amastech.cloud/authorize',headers={'Content-Type': 'application/json'}, json={'refresh_token': refresh_token, 'identity_token': identity_token, 'timestamp': timestamp}, verify=certifi.where())
                if req_resp:
                    print('Validated!')
                    updated_tokens = req_resp.json()
                    for i in updated_tokens:
                        cookie_jar.put(i, value=encrypt(str(updated_tokens[i]).encode(), key))
                    identity_token = updated_tokens['identity_token']
                    setup_dict = {'id_token': identity_token}
                else:
                    print('NOT validated, needs to sign in again!')
                    setup_dict = {}
            except:
                setup_dict = {}
        setup_serial = self.ids.serial_txt.text
        print('Got serial: '+setup_serial)
        setup_api_key = self.ids.api_txt.text
        print('Got API key too')
        ssid = self.ids.ssid_txt.text
        print('Setting up in WiFi: '+ssid)
        wifi_password = self.ids.wifi_pass_txt.text
        print('WiFi is already secured with a password')
        registration = self.ids.registration_checkbox.state
        print('Registration to user account is: '+registration)
        optional_name = self.ids.name_txt.text
        if optional_name:
            print('Got friendly name: '+optional_name)
        optional_ip = self.ids.ip_txt.text
        if optional_ip:
            print('Got optional IP: '+optional_ip)

        if optional_ip and optional_ip != '192.168.4.1' and registration == 'down' and not 'local_only' in self.cookie_jar.keys():            
            if optional_name:
                setup_dict.update({'friendly_name': optional_name})
            try:
                resp = requests.post('https://'+optional_ip+'/setup', headers={'x-api-key': setup_api_key, 'Content-Type': 'application/json', 'Accept': 'application/json'}, json=setup_dict, verify=base_dir+'assets/static/AMAS_CA.pem')
            except: resp = None
        elif optional_ip and optional_ip != '192.168.4.1' and ('local_only' in cookie_jar.keys() or registration == 'normal'):
            print('Device already locally connected in your network, please add it as an existing device!')
            return False
        elif not optional_ip or optional_ip == '192.168.4.1':
            if ssid and wifi_password and registration == 'down' and not 'local_only' in cookie_jar.keys():
                if refresh_token:
                    print('Validating existing session...')
                    try:
                        req_resp = requests.post(url='https://amastech.cloud/authorize',headers={'Content-Type': 'application/json'}, json={'refresh_token': refresh_token, 'identity_token': identity_token, 'timestamp': timestamp}, verify=certifi.where())
                        if req_resp:
                            print('Validated!')
                            updated_tokens = req_resp.json()
                            for i in updated_tokens:
                                cookie_jar.put(i, value=encrypt(str(updated_tokens[i]).encode(), key))
                            identity_token = updated_tokens['identity_token']
                            setup_dict = {'id_token': identity_token}
                        else:
                            print('NOT validated, needs to sign in again!')
                            setup_dict = {}
                    except:
                        setup_dict = {}
                setup_dict.update({'ssid': ssid, 'password': wifi_password})
                if optional_name:
                    setup_dict.update({'friendly_name': optional_name})
                self.dismiss_loading()
                self.confirmation_dialog(setup_dict, setup_api_key)
                return True
            elif ssid and wifi_password and ('local_only' in cookie_jar.keys() or registration == 'normal'):
                setup_dict = {'ssid': ssid, 'password': wifi_password}
                if optional_name:
                    setup_dict.update({'friendly_name': optional_name})
                self.dismiss_loading()
                self.confirmation_dialog(setup_dict, setup_api_key)
                return True
            else:
                print('Both WiFi name and the WiFi password is required!')
                return False
        else:
            print('Unforeseen circumstance, please inform the developers!!')
        
        if resp:
            if optional_ip and optional_ip != '192.168.4.1':
                print('Successfully registered the local device with your account! Redirecting to your dashboard after a short wait...')
            else:
                print('Successfully set up the new device! Redirecting to your dashboard after a short wait...')
            self.redirect_screen()
        elif resp is not None and resp.status_code == 401:
            print('Wrong API key!')
        else:
            print('Error setting up the device!')
        self.dismiss_loading()


    @mainthread
    def confirmation_dialog(self, setup_dict, setup_api_key):
        confirmation_modal = MDDialog(
            MDDialogIcon(
                icon='wifi-cog'
            ),
            MDDialogHeadlineText(
                text='Are you connected to your device?'
            ),
            MDDialogSupportingText(
                text='Before we can proceed, you need to be connected to your device directly. You can do that by connecting to your device\'s WiFi access point network as noted below.'
            ),
        )
        confirmation_modal.add_widget(MDDialogContentContainer(
                MDDivider(),
                MDBoxLayout(
                    MDBoxLayout(
                        size_hint_y=None,
                        height='15dp',
                    ),
                    MDFloatLayout(
                        MDLabel(
                            text="WiFi Name (SSID):",
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.9},
                            padding='12dp',
                        ),
                        MDLabel(
                            text='amas'+self.ids.serial_txt.text.lower(),
                            allow_selection=True,
                            allow_copy=True,
                            adaptive_size=True,
                            pos_hint={'right': 1, 'center_y': 0.9},
                            padding='8dp',
                        ),
                    ),
                    MDBoxLayout(
                        size_hint_y=None,
                        height='15dp',
                    ),
                    MDFloatLayout(
                        MDLabel(
                            text="WiFi Password (WPA2-PSK):",
                            adaptive_size=True,
                            pos_hint={'left': 1, 'center_y': 0.1},
                            padding='12dp',
                        ),
                        MDLabel(
                            text=self.ids.api_txt.text,
                            allow_selection=True,
                            allow_copy=True,
                            adaptive_size=True,
                            pos_hint={'right': 1, 'center_y': 0.1},
                            padding='8dp',
                        ),
                    ),
                    MDBoxLayout(
                        size_hint_y=None,
                        height='15dp',
                    ),
                    orientation='vertical',
                    size_hint_y=None,
                    height='80dp',
                ),
                MDDivider(),
                orientation="vertical"
            ),
        )
        confirmation_modal.add_widget(MDDialogButtonContainer(
                Widget(),
                MDButton(
                    MDButtonText(text="Cancel"),
                    style="text",
                    on_release=lambda *args: self.cancel_confirm_dialog(confirmation_modal),
                ),
                MDButton(
                    MDButtonText(text="Proceed"),
                    style="text",
                    on_release=lambda *args: self.load_proceed_setup(confirmation_modal, setup_dict, setup_api_key),
                ),
                spacing="8dp",
            )
        )
        confirmation_modal.open()


    @mainthread
    def cancel_confirm_dialog(self, confirmation_modal: MDDialog):
        confirmation_modal.dismiss()


    @mainthread
    def load_proceed_setup(self, confirmation_modal: MDDialog, setup_dict: dict, setup_api_key: str):
        confirmation_modal.dismiss()
        self.show_loading()
        Thread(target=lambda *args: self.proceed_setup(confirmation_modal, setup_dict, setup_api_key), daemon=True, name='setup-device').start()


    def proceed_setup(self, confirmation_modal: MDDialog, setup_dict: dict, setup_api_key: str):
        try:
            resp = requests.post('https://192.168.4.1/setup', headers={'x-api-key': setup_api_key, 'Content-Type': 'application/json', 'Accept': 'application/json'}, json=setup_dict, verify=base_dir+'assets/static/AMAS_CA.pem')
        except: resp = None

        if resp:
            print('Successfully set up the new device! Redirecting to your dashboard after a short wait...')
            self.redirect_screen()
        elif resp is not None and resp.status_code == 401:
            print('Wrong API key!')
        else:
            print('Error setting up the device!')
        self.dismiss_loading()


    @mainthread
    def redirect_screen(self, screen = 'dashboard'):
        self.hamburger_menu.dismiss()
        self.dismiss_loading()
        self.manager.current = screen
