from kivymd.uix.screen import MDScreen
import webbrowser
from urllib.parse import quote_plus
from threading import Thread
import requests, certifi
from kivy.storage.jsonstore import JsonStore
from aiohttp import web
import asyncio
from assets.env import *
from json import loads
from time import time
from kivy.clock import Clock


class UserChoiceScreen(MDScreen):
    global flask_host, flask_port
    def __init__(self, **kwargs):
        super(UserChoiceScreen, self).__init__(**kwargs)

        self.client_id = None
        self.response_type = None
        self.scope = None
        self.redirect_uri = None
        self.url = None
        self.session_uri = None
        self.refresh_uri = None
        self.cookie_jar = JsonStore(data_path+'cookies.json')
        self.aiohttp_app = None

        Thread(target=self.start_aiohttp, daemon=True, name='aiohttp-server').start()

    def aiohttp_server(self):
        self.aiohttp_app = web.Application(debug=True)

        def index(request: web.Request):
            settings_store = JsonStore(data_path+'settings.json')
            flask_host = settings_store.get('flask')['host']
            flask_port = settings_store.get('flask')['port']
            self.client_id = '3113cl3d4frk5cgrquus56j81a'
            self.response_type = 'code'
            self.scope = 'openid+profile+email+phone'
            self.redirect_uri = quote_plus('http://'+flask_host+':'+str(flask_port))
            url = 'https://auth.amastech.cloud/oauth2/authorize?redirect_uri={redirect_uri}&response_type={response_type}&scope={scope}&client_id={client_id}'
            self.url = url.format(redirect_uri=self.redirect_uri, response_type=self.response_type, scope=self.scope, client_id=self.client_id)
            self.session_uri = 'https://auth.amastech.cloud/oauth2/token?grant_type=authorization_code&client_id={client_id}&redirect_uri={redirect_uri}&code={code}'
            self.refresh_uri = 'https://auth.amastech.cloud/oauth2/token?grant_type=refresh_token&client_id={client_id}&redirect_uri={redirect_uri}&refresh_token={refresh_token}'
            if 'code' in request.rel_url.query:
                print('Got authorization code...')
                authorization_code = request.rel_url.query['code']
                session_uri_f = self.session_uri.format(client_id=self.client_id, redirect_uri=self.redirect_uri, code=authorization_code)
                session = requests.post(url=session_uri_f, headers={'Content-type': 'application/x-www-form-urlencoded'}, verify=certifi.where())
                if session:
                    print('Session authenticated...')
                    session = loads(session.content)
                    identity_token = str(session['id_token'])
                    refresh_token = str(session['refresh_token'])
                    timestamp = str(int(time()))
                    cookie_jar = JsonStore(data_path+'cookies.json')
                    cookie_jar.put('identity_token', value=encrypt(identity_token.encode(), key))
                    cookie_jar.put('refresh_token', value=encrypt(refresh_token.encode(), key))
                    cookie_jar.put('timestamp', value=encrypt(timestamp.encode(), key))
                    if 'wait_for_validation' in cookie_jar.keys():
                        cookie_jar.delete('wait_for_validation')
                    # self.loop.create_task(self.redirect_screen())
                    Clock.schedule_once(self.redirect_screen)
                    return web.Response(body='You can now safely close this page', status=200)
                else:
                    print('Redirecting to authorization')
                    return web.HTTPFound(self.url)
            elif 'refresh_token' in request.rel_url.query:
                print('Got refresh token, refreshing session...')
                refresh_token = request.rel_url.query['refresh_token']
                refresh_uri_f = self.refresh_uri.format(client_id=self.client_id, redirect_uri=self.redirect_uri, refresh_token=refresh_token)
                session_refresh = requests.post(url=refresh_uri_f, headers={'Content-type': 'application/x-www-form-urlencoded'}, verify=certifi.where())
                if session_refresh:
                    print('Session authenticated...')
                    session = loads(session.content)
                    identity_token = str(session['id_token'])
                    timestamp = str(int(time()))
                    cookie_jar = JsonStore(data_path+'cookies.json')
                    cookie_jar.put('identity_token', value=encrypt(identity_token.encode(), key))
                    cookie_jar.put('timestamp', value=encrypt(timestamp.encode(), key))
                    if 'wait_for_validation' in cookie_jar.keys():
                        cookie_jar.delete('wait_for_validation')
                    return web.Response(body='You can now safely close this page', status=200)
            else:
                print('Redirecting to authorization')
                return web.HTTPFound(self.url)
        
        self.aiohttp_app.add_routes([web.get('/', index),
                                     web.post('/', index)])
        handler = self.aiohttp_app.make_handler()
    
        return handler
    
    def start_aiohttp(self):
        loop = asyncio.get_event_loop_policy().new_event_loop()
        asyncio.set_event_loop(loop)

        handler = self.aiohttp_server()

        settings_store = JsonStore(data_path+'settings.json')
        flask_host = settings_store.get('flask')['host']
        flask_port = settings_store.get('flask')['port']
        try_direction = 'up'

        try:
            server = loop.create_server(handler, host=flask_host, port=flask_port)
            loop.run_until_complete(server)
            loop.run_forever()
        except:
            try:
                if flask_port+1 in range(16457,16467):
                    flask_port += 1
                    try_direction = 'up'
                else:
                    flask_port -= 1
                    try_direction = 'down'
                server = loop.create_server(handler, host=flask_host, port=flask_port)
                loop.run_until_complete(server)
                settings_store.put('flask', host=flask_host, port=flask_port)
                loop.run_forever()
            except:
                try:
                    if try_direction == 'up':
                        if flask_port+1 in range(16457,16467):
                            flask_port += 1
                        else:
                            flask_port -= 1
                    else:
                        if flask_port-1 in range(16457,16467):
                            flask_port -= 1
                        else:
                            flask_port += 1
                    server = loop.create_server(handler, host=flask_host, port=flask_port)
                    loop.run_until_complete(server)
                    settings_store.put('flask', host=flask_host, port=flask_port)
                    loop.run_forever()
                except:
                    try:
                        if try_direction == 'up':
                            if flask_port+1 in range(16457,16467):
                                flask_port += 1
                            else:
                                flask_port -= 1
                        else:
                            if flask_port-1 in range(16457,16467):
                                flask_port -= 1
                            else:
                                flask_port += 1
                        server = loop.create_server(handler, host=flask_host, port=flask_port)
                        loop.run_until_complete(server)
                        settings_store.put('flask', host=flask_host, port=flask_port)
                        loop.run_forever()
                    except:
                        try:
                            if try_direction == 'up':
                                if flask_port+1 in range(16457,16467):
                                    flask_port += 1
                                else:
                                    flask_port -= 1
                            else:
                                if flask_port-1 in range(16457,16467):
                                    flask_port -= 1
                                else:
                                    flask_port += 1
                            server = loop.create_server(handler, host=flask_host, port=flask_port)
                            loop.run_until_complete(server)
                            settings_store.put('flask', host=flask_host, port=flask_port)
                            loop.run_forever()
                        except:
                            if try_direction == 'up':
                                if flask_port+1 in range(16457,16467):
                                    flask_port += 1
                                else:
                                    flask_port -= 1
                            else:
                                if flask_port-1 in range(16457,16467):
                                    flask_port -= 1
                                else:
                                    flask_port += 1
                            server = loop.create_server(handler, host=flask_host, port=flask_port)
                            loop.run_until_complete(server)
                            settings_store.put('flask', host=flask_host, port=flask_port)
                            loop.run_forever()

    def login(self):
        settings_store = JsonStore(data_path+'settings.json')
        flask_host = settings_store.get('flask')['host']
        flask_port = settings_store.get('flask')['port']
        webbrowser.open('http://'+flask_host+':'+str(flask_port))

    def local_only_setup(self):
        print('App set up as local only, redirecting to dashbard...')
        self.cookie_jar = JsonStore(data_path+'cookies.json')
        self.cookie_jar.put('local_only', value=True)
        if 'wait_for_validation' in self.cookie_jar.keys():
                self.cookie_jar.delete('wait_for_validation')
        self.manager.current = 'dashboard'

    def redirect_screen(self, *args):
        self.manager.current = 'dashboard'